# pylint: disable=no-name-in-module
# pylint: disable=too-many-lines
import logging
import warnings

from copy import deepcopy

import numpy as np
import pandas as pd
import pymc as pm

from arviz.plots import plot_posterior

from formulae import design_matrices

from .backend import PyMCModel
from .defaults import get_default_prior, get_builtin_family
from .families import Family, univariate, multivariate
from .priors import Prior, PriorScaler
from .terms import GroupSpecificTerm, OffsetTerm, ResponseTerm, Term
from .utils import listify, extra_namespace
from .version import __version__

_log = logging.getLogger("bambi")


class Model:
    """Specification of model class.

    Parameters
    ----------
    formula : str
        A model description written using the formula syntax from the ``formulae`` library.
    data : pandas.DataFrame or str
        The dataset to use. Either a pandas ``DataFrame``, or the name of the file containing
        the data, which will be passed to ``pd.read_csv()``.
    family : str or bambi.families.Family
        A specification of the model family (analogous to the family object in R). Either
        a string, or an instance of class ``bambi.families.Family``. If a string is passed, a
        family with the corresponding name must be defined in the defaults loaded at ``Model``
        initialization. Valid pre-defined families are ``"bernoulli"``, ``"beta"``,
        ``"binomial"``, ``"categorical"``, ``"gamma"``, ``"gaussian"``, ``"negativebinomial"``,
        ``"poisson"``, ``"t"``, and ``"wald"``. Defaults to ``"gaussian"``.
    priors : dict
        Optional specification of priors for one or more terms. A dictionary where the keys are
        the names of terms in the model, "common" or "group_specific" and the values are
        instances of class ``Prior`` when ``automatic_priors`` is ``"default"``.
    link : str
        The name of the link function to use. Valid names are ``"cloglog"``, ``"identity"``,
        ``"inverse_squared"``, ``"inverse"``, ``"log"``, ``"logit"``, ``"probit"``, and
        ``"softmax"``. Not all the link functions can be used with all the families.
    categorical : str or list
        The names of any variables to treat as categorical. Can be either a single variable
        name, or a list of names. If categorical is ``None``, the data type of the columns in
        the ``data`` will be used to infer handling. In cases where numeric columns are
        to be treated as categorical (e.g., group specific factors coded as numerical IDs),
        explicitly passing variable names via this argument is recommended.
    potentials : A list of 2-tuples.
        Optional specification of potentials. A potential is an arbitrary expression added to the
        likelihood, this is generally useful to add constrains to models, that are difficult to
        express otherwise. The first term of a 2-tuple is the name of a variable in the model, the
        second a lambda function expressing the desired constraint.
        If a constraint involves n variables, you can pass n 2-tuples or pass a tuple which first
        element is a n-tuple and second element is a lambda function with n arguments. The number
        and order of the lambda function has to match the number and order of the variables names.
    dropna : bool
        When ``True``, rows with any missing values in either the predictors or outcome are
        automatically dropped from the dataset in a listwise manner.
    auto_scale : bool
        If ``True`` (default), priors are automatically rescaled to the data
        (to be weakly informative) any time default priors are used. Note that any priors
        explicitly set by the user will always take precedence over default priors.
    automatic_priors: str
        An optional specification to compute automatic priors. ``"default"`` means to use
        a method inspired on the R rstanarm library.
    noncentered : bool
        If ``True`` (default), uses a non-centered parameterization for normal hyperpriors on
        grouped parameters. If ``False``, naive (centered) parameterization is used.
    priors_cor : dict
        An optional value for eta in the LKJ prior for the correlation matrix of group-specific
        terms. Keys in the dictionary indicate the groups, and values indicate the value of eta.
        This is a very experimental feature. Defaults to ``None``, which means priors for the
        group-specific terms are independent.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        formula,
        data,
        family="gaussian",
        priors=None,
        link=None,
        categorical=None,
        potentials=None,
        dropna=False,
        auto_scale=True,
        automatic_priors="default",
        noncentered=True,
        priors_cor=None,
    ):
        # attributes that are set later
        self.terms = {}
        self.built = False  # build()
        self._backend_name = None

        # build() will loop over this, calling _set_priors()
        self._added_priors = {}

        self._design = None
        self.formula = None
        self.response = None  # _add_response()
        self.family = None  # _add_response()
        self.backend = None  # _set_backend()
        self.priors_cor = {}  # _add_priors_cor()

        self.auto_scale = auto_scale
        self.dropna = dropna
        self.noncentered = noncentered
        self.potentials = potentials

        # Read and clean data
        if isinstance(data, str):
            data = pd.read_csv(data, sep=None, engine="python")
        elif not isinstance(data, pd.DataFrame):
            raise ValueError("'data' must be a string with a path to a .csv or a pandas DataFrame.")

        # Convert 'object' and explicitly asked columns to categorical.
        object_cols = list(data.select_dtypes("object").columns)
        cols_to_convert = list(set(object_cols + listify(categorical)))
        if cols_to_convert:
            data = data.copy()  # don't modify original data frame
            data[cols_to_convert] = data[cols_to_convert].apply(lambda x: x.astype("category"))

        self.data = data

        # Handle priors
        if priors is None:
            priors = {}
        else:
            priors = deepcopy(priors)

        self.automatic_priors = automatic_priors

        # Obtain design matrices and related objects.
        na_action = "drop" if dropna else "error"
        self.formula = formula
        self._design = design_matrices(formula, data, na_action, 1, extra_namespace)

        if self._design.response is None:
            raise ValueError(
                "No outcome variable is set! "
                "Please specify an outcome variable using the formula interface."
            )

        self._set_family(family, link, priors)
        self._add_response(self._design.response)

        if self._design.common:
            self._add_common(self._design.common, priors)

        if self._design.group:
            self._add_group_specific(self._design.group, priors)

        if priors_cor:
            self._add_priors_cor(priors_cor)

        # Build priors
        self._build_priors()

    def fit(
        self,
        draws=1000,
        tune=1000,
        discard_tuned_samples=True,
        omit_offsets=True,
        include_mean=False,
        inference_method="mcmc",
        init="auto",
        n_init=50000,
        chains=None,
        cores=None,
        random_seed=None,
        **kwargs,
    ):
        """Fit the model using PyMC.

        Parameters
        ----------
        draws: int
            The number of samples to draw from the posterior distribution. Defaults to 1000.
        tune : int
            Number of iterations to tune. Defaults to 1000. Samplers adjust the step sizes,
            scalings or similar during tuning. These tuning samples are be drawn in addition to the
            number specified in the ``draws`` argument, and will be discarded unless
            ``discard_tuned_samples`` is set to ``False``.
        discard_tuned_samples : bool
            Whether to discard posterior samples of the tune interval. Defaults to ``True``.
        omit_offsets: bool
            Omits offset terms in the ``InferenceData`` object returned when the model includes
            group specific effects. Defaults to ``True``.
        include_mean: bool
            Compute the posterior of the mean response. Defaults to ``False``.
        inference_method: str
            The method to use for fitting the model. By default, ``"mcmc"``. This automatically
            assigns a MCMC method best suited for each kind of variables, like NUTS for continuous
            variables and Metropolis for non-binary discrete ones. Alternatively, ``"vi"``, in
            which case the model will be fitted using variational inference as implemented in PyMC
            using the ``fit`` function.
            Finally, ``"laplace"``, in which case a Laplace approximation is used and is not
            recommended other than for pedagogical use.
            To use the PyMC numpyro and blackjax samplers, use ``nuts_numpyro`` or ``nuts_blackjax``
            respectively. Both methods will only work if you can use NUTS sampling, so your model
            must be differentiable.
        init: str
            Initialization method. Defaults to ``"auto"``. The available methods are:
            * auto: Use ``"jitter+adapt_diag"`` and if this method fails it uses ``"adapt_diag"``.
            * adapt_diag: Start with a identity mass matrix and then adapt a diagonal based on the
            variance of the tuning samples. All chains use the test value (usually the prior mean)
            as starting point.
            * jitter+adapt_diag: Same as ``"adapt_diag"``, but use test value plus a uniform jitter
            in [-1, 1] as starting point in each chain.
            * advi+adapt_diag: Run ADVI and then adapt the resulting diagonal mass matrix based on
            the sample variance of the tuning samples.
            * advi+adapt_diag_grad: Run ADVI and then adapt the resulting diagonal mass matrix based
            on the variance of the gradients during tuning. This is **experimental** and might be
            removed in a future release.
            * advi: Run ADVI to estimate posterior mean and diagonal mass matrix.
            * advi_map: Initialize ADVI with MAP and use MAP as starting point.
            * map: Use the MAP as starting point. This is strongly discouraged.
            * adapt_full: Adapt a dense mass matrix using the sample covariances. All chains use the
            test value (usually the prior mean) as starting point.
            * jitter+adapt_full: Same as ``"adapt_full"``, but use test value plus a uniform jitter
            in [-1, 1] as starting point in each chain.
        n_init: int
            Number of initialization iterations. Only works for ``"advi"`` init methods.
        chains: int
            The number of chains to sample. Running independent chains is important for some
            convergence statistics and can also reveal multiple modes in the posterior. If ``None``,
            then set to either ``cores`` or 2, whichever is larger.
        cores : int
            The number of chains to run in parallel. If ``None``, it is equal to the number of CPUs
            in the system unless there are more than 4 CPUs, in which case it is set to 4.
        random_seed : int or list of ints
            A list is accepted if cores is greater than one.
        **kwargs:
            For other kwargs see the documentation for ``PyMC.sample()``.

        Returns
        -------
        An ArviZ ``InferenceData`` instance if inference_method is  ``"mcmc"`` (default),
        "nuts_numpyro", "nuts_blackjax" or "laplace".
        An ``Approximation`` object if  ``"vi"``.
        """
        method = kwargs.pop("method", None)
        if method is not None:
            if inference_method == "vi":
                kwargs["method"] = method
            else:
                warnings.warn(
                    "the method argument has been deprecated, please use inference_method",
                    FutureWarning,
                )
                inference_method = method

        if not self.built:
            self.build()

        # Tell user which event is being modeled
        if isinstance(self.family, univariate.Bernoulli):
            _log.info(
                "Modeling the probability that %s==%s",
                self.response.name,
                str(self.response.success),
            )

        return self.backend.run(
            draws=draws,
            tune=tune,
            discard_tuned_samples=discard_tuned_samples,
            omit_offsets=omit_offsets,
            include_mean=include_mean,
            inference_method=inference_method,
            init=init,
            n_init=n_init,
            chains=chains,
            cores=cores,
            random_seed=random_seed,
            **kwargs,
        )

    def build(self):
        """Set up the model for sampling/fitting.

        Creates an instance of the underlying PyMC model and adds all the necessary terms to it.
        """
        self.backend = PyMCModel()
        self.backend.build(self)
        self.built = True

    def set_priors(self, priors=None, common=None, group_specific=None):
        """Set priors for one or more existing terms.

        Parameters
        ----------
        priors: dict
            Dictionary of priors to update. Keys are names of terms to update; values are the new
            priors (either a ``Prior`` instance, or an int or float that scales the default priors).
            Note that a tuple can be passed as the key, in which case the same prior will be applied
            to all terms named in the tuple.
        common: Prior, int, float or str
            A prior specification to apply to all common terms included in the model.
        group_specific: Prior, int, float or str
            A prior specification to apply to all group specific terms included in the model.
        """
        kwargs = dict(zip(["priors", "common", "group_specific"], [priors, common, group_specific]))
        self._added_priors.update(kwargs)
        # After updating, we need to rebuild priors.
        self._build_priors()
        self.built = False

    def _build_priors(self):
        """Carry out all operations related to the construction and/or scaling of priors."""
        # Set custom priors that have been passed via `Model.set_priors()`
        self._set_priors(**self._added_priors)

        # Prepare all priors
        for term in self.terms.values():
            if term.group_specific:
                kind = "group_specific"
            elif term.kind == "intercept":
                kind = "intercept"
            elif term.kind == "offset":
                continue
            else:
                kind = "common"
            term.prior = prepare_prior(term.prior, kind, self.auto_scale)

        # Scale priors if there is at least one term in the model and auto_scale is True
        if self.terms and self.auto_scale:
            method = self.automatic_priors
            if method == "default":
                scaler = PriorScaler(self)
            else:
                raise ValueError(
                    f"{method} is not a valid method for default priors. Use 'default'."
                )
            self.scaler = scaler
            self.scaler.scale()

    def _set_priors(self, priors=None, common=None, group_specific=None):
        """Internal version of ``set_priors()``, with same arguments.

        Runs during ``Model._build_priors()``.
        """
        # First, it constructs a `targets` dict where it store key-value (name-prior) pairs that
        # are going to be updated. Finally, the update is done in the last for loop in this method.
        targets = {}

        if common is not None:
            targets.update({name: common for name in self.common_terms})

        if group_specific is not None:
            targets.update({name: group_specific for name in self.group_specific_terms})

        if priors is not None:
            priors = deepcopy(priors)
            # Prepare priors for auxiliary parameters of the response
            family_prior = {
                name: priors.pop(name) for name in self.family.likelihood.priors if name in priors
            }
            if family_prior:
                for prior in family_prior.values():
                    prior.auto_scale = False
                self.family.likelihood.priors.update(family_prior)

            # Prepare priors for explanatory terms.
            for names, prior in priors.items():
                # In case we have tuple-keys, we loop through each of them.
                for name in listify(names):
                    if name not in self.terms:
                        raise ValueError(f"No terms in model match '{name}'.")
                    targets[name] = prior

        # Set priors for explanatory terms.
        for name, prior in targets.items():
            self.terms[name].prior = prior

    def _set_family(self, family, link, priors):
        """Set the Family of the model.

        Parameters
        ----------
        family: str or bambi.families.Family
            A specification of the model family (analogous to the family object in R). Either a
            string, or an instance of class ``families.Family``. If a string is passed, a family
            with the corresponding name must be defined in the defaults loaded at model
            initialization.
        link: str
            The name of the link function to use. Valid names are ``"cloglog"``, ``"identity"``,
            ``"inverse_squared"``, ``"inverse"``, ``"log"``, ``"logit"``, ``"probit"``, and
            ``"softmax"``. Not all the link functions can be used with all the families.
        priors: dict
            A dictionary with specification of priors for the parameters in the family of
            the response. Keys are names of other parameters than the mean in the family
            (i.e. they cannot be equal to ``family.parent``) and values can be an instance of class
            ``Prior``, a numeric value, or a string describing the width. In the numeric case,
            the distribution specified in the defaults will be used, and the passed value will be
            used to scale the appropriate variance parameter.
        """

        # If string, get builtin family
        if isinstance(family, str):
            family = get_builtin_family(family)

        # Always ensure family is indeed instance of Family
        if not isinstance(family, Family):
            raise ValueError("'family' must be a string or a Family object.")

        # Get the names of the auxiliary parameters in the family
        aux_params = list(family.likelihood.priors)

        # Check if any of the names of the auxiliary params match the names of terms in the model
        # If that happens, raise an error.
        if aux_params and self._design.common:
            conflicts = [name for name in aux_params if name in self._design.common.terms]
            if conflicts:
                raise ValueError(
                    f"The prior name for {', '.join(conflicts)} conflicts with the name of a "
                    "parameter in the response distribution.\n"
                    "Please rename the term(s) to prevent an unexpected behaviour."
                )

        # Extract priors for auxiliary params
        priors = {k: v for k, v in priors.items() if k in aux_params}

        # Update auxiliary parameters
        if priors:
            for prior in priors.values():
                if isinstance(prior, Prior):
                    prior.auto_scale = False
            family.likelihood.priors.update(priors)

        # Override family's link if another is explicitly passed
        if link is not None:
            family.link = link

        self.family = family

    def set_alias(self, aliases):
        """Set aliases for the terms and auxiliary parameters in the model

        Parameters
        ----------
        aliases: dict
            A dictionary where key represents the original term name and the value is the alias.
        """
        if not isinstance(aliases, dict):
            raise ValueError(f"'aliases' must be a dictionary, not a {type(aliases)}.")

        for name, alias in aliases.items():
            if name in self.terms:
                self.terms[name].set_alias(alias)
            if name in self.family.likelihood.priors:
                self.family.set_alias(name, alias)
            if name == self.response.name:
                self.response.set_alias(alias)

            # Now add aliases for hyperpriors in group specific terms
            for term in self.group_specific_terms.values():
                if name in term.prior.args:
                    term.set_hyperprior_alias(name, alias)

        # Model needs to be rebuilt after modifying aliases
        self.built = False

    def _add_response(self, response):
        """Add a response (or outcome/dependent) variable to the model.

        Parameters
        ----------
        response : formulae.ResponseMatrix
            An instance of ``formulae.ResponseMatrix`` as returned by
            ``formulae.design_matrices()``.
        """

        if hasattr(response.term.term.components[0], "reference"):
            reference = response.term.term.components[0].reference
        else:
            reference = None

        if reference is not None and not isinstance(self.family, univariate.Bernoulli):
            raise ValueError("Index notation for response is only available for 'bernoulli' family")

        if isinstance(self.family, univariate.Bernoulli):
            if response.kind == "categoric" and response.levels is None and reference is None:
                raise ValueError("Categoric response must be binary for 'bernoulli' family.")
            if response.kind == "numeric" and not all(np.isin(response.design_matrix, [0, 1])):
                raise ValueError("Numeric response must be all 0 and 1 for 'bernoulli' family.")

        self.response = ResponseTerm(response, self)
        self.built = False

    def _add_common(self, common, priors):
        """Add common (a.k.a. fixed) terms to the model.

        Parameters
        ----------
        common: formulae.CommonEffectsMatrix
            Representation of the design matrix for the common effects of a model. It contains all
            the necessary information to build the ``Term`` objects associated with each common
            term in the model.
        priors: dict
            Optional specification of priors for one or more terms. A dictionary where the keys are
            any of the names of the common terms in the model or ``"common"`` and the values are
            either instances of class ``Prior`` or ``int``, ``float``, or ``str`` that specify the
            width of the priors on a standardized scale.
        """
        for name, term in common.terms.items():
            data = common[name]
            prior = priors.pop(name, priors.get("common", None))
            if isinstance(prior, Prior):
                any_hyperprior = any(isinstance(x, Prior) for x in prior.args.values())
                if any_hyperprior:
                    raise ValueError(
                        f"Trying to set hyperprior on '{name}'. "
                        "Can't set a hyperprior on common effects."
                    )
            if term.kind == "offset":
                self.terms[name] = OffsetTerm(name, term, data)
            else:
                self.terms[name] = Term(name, term, data, prior)

    def _add_group_specific(self, group, priors):
        """Add group-specific (a.k.a. random) terms to the model.

        Parameters
        ----------
        group: formulae.GroupEffectsMatrix
            Representation of the design matrix for the group specific effects of a model. It
            contains all the necessary information to build the ``GroupSpecificTerm`` objects
            associated with each group-specific term in the model.
        priors: dict
            Optional specification of priors for one or more terms. A dictionary where the keys are
            any of the names of the group-specific terms in the model or ``"group_specific"`` and
            the values are either instances of class ``Prior`` or ``int``, ``float``, or ``str``
            that specify the width of the priors on a standardized scale.
        """
        for name, term in group.terms.items():
            data = group[name]
            prior = priors.pop(name, priors.get("group_specific", None))
            self.terms[name] = GroupSpecificTerm(name, term, data, prior)

    def _add_priors_cor(self, priors):
        # priors: dictionary. names are groups, values are the "eta" in the lkj prior
        groups = self._get_group_specific_groups()
        for group in groups:
            if group in priors:
                self.priors_cor[group] = priors[group]
            else:
                raise KeyError(f"The name {group} is not a group in any group-specific term.")

    def plot_priors(
        self,
        draws=5000,
        var_names=None,
        random_seed=None,
        figsize=None,
        textsize=None,
        hdi_prob=None,
        round_to=2,
        point_estimate="mean",
        kind="kde",
        bins=None,
        omit_offsets=True,
        omit_group_specific=True,
        ax=None,
    ):
        """
        Samples from the prior distribution and plots its marginals.

        Parameters
        ----------
        draws: int
            Number of draws to sample from the prior predictive distribution. Defaults to 5000.
        var_names: str or list
            A list of names of variables for which to compute the posterior predictive
            distribution. Defaults to ``None`` which means to include both observed and
            unobserved RVs.
        random_seed: int
            Seed for the random number generator.
        figsize: tuple
            Figure size. If ``None`` it will be defined automatically.
        textsize: float
            Text size scaling factor for labels, titles and lines. If ``None`` it will be
            autoscaled based on ``figsize``.
        hdi_prob: float or str
            Plots highest density interval for chosen percentage of density.
            Use ``"hide"`` to hide the highest density interval. Defaults to 0.94.
        round_to: int
            Controls formatting of floats. Defaults to 2 or the integer part, whichever is bigger.
        point_estimate: str
            Plot point estimate per variable. Values should be ``"mean"``, ``"median"``, ``"mode"``
            or ``None``. Defaults to ``"auto"`` i.e. it falls back to default set in
            ArviZ's rcParams.
        kind: str
            Type of plot to display (``"kde"`` or ``"hist"``) For discrete variables this argument
            is ignored and a histogram is always used.
        bins: integer or sequence or "auto"
            Controls the number of bins, accepts the same keywords ``matplotlib.pyplot.hist()``
            does. Only works if ``kind == "hist"``. If ``None`` (default) it will use ``"auto"``
            for continuous variables and ``range(xmin, xmax + 1)`` for discrete variables.
        omit_offsets: bool
            Whether to omit offset terms in the plot. Defaults to ``True``.
        omit_group_specific: bool
            Whether to omit group specific effects in the plot. Defaults to ``True``.
        ax: numpy array-like of matplotlib axes or bokeh figures
            A 2D array of locations into which to plot the densities. If not supplied, ArviZ will
            create its own array of plot areas (and return it).
        **kwargs
            Passed as-is to ``matplotlib.pyplot.hist()`` or ``matplotlib.pyplot.plot()`` function
            depending on the value of ``kind``.

        Returns
        -------
        axes: matplotlib axes
        """
        if not self.built:
            raise ValueError(
                "Cannot plot priors until model is built!! "
                "Call .build() to build the model or .fit() to build and sample from the posterior."
            )

        unobserved_rvs_names = []
        flat_rvs = []
        for unobserved in self.backend.model.unobserved_RVs:
            if "Flat" in unobserved.__str__():
                flat_rvs.append(unobserved.name)
            else:
                unobserved_rvs_names.append(unobserved.name)
        if var_names is None:
            var_names = pm.util.get_default_varnames(
                unobserved_rvs_names, include_transformed=False
            )
        else:
            flat_rvs = [fv for fv in flat_rvs if fv in var_names]
            var_names = [vn for vn in var_names if vn not in flat_rvs]

        if flat_rvs:
            _log.info(
                "Variables %s have flat priors, and hence they are not plotted", ", ".join(flat_rvs)
            )

        if omit_offsets:
            var_names = [name for name in var_names if not name.endswith("_offset")]

        if omit_group_specific:
            var_names = [vn for vn in var_names if vn not in self.group_specific_terms]

        axes = None
        if var_names:
            # Sort variable names so Intercept is in the beginning
            if "Intercept" in var_names:
                var_names.insert(0, var_names.pop(var_names.index("Intercept")))
            pps = self.prior_predictive(draws=draws, var_names=var_names, random_seed=random_seed)

            axes = plot_posterior(
                pps,
                group="prior",
                var_names=var_names,
                figsize=figsize,
                textsize=textsize,
                hdi_prob=hdi_prob,
                round_to=round_to,
                point_estimate=point_estimate,
                kind=kind,
                bins=bins,
                ax=ax,
            )
        return axes

    def prior_predictive(self, draws=500, var_names=None, omit_offsets=True, random_seed=None):
        """
        Generate samples from the prior predictive distribution.
        Parameters
        ----------
        draws: int
            Number of draws to sample from the prior predictive distribution. Defaults to 500.
        var_names: str or list
            A list of names of variables for which to compute the prior predictive distribution.
            Defaults to ``None`` which means both observed and unobserved RVs.
        random_seed: int
            Seed for the random number generator.
        Returns
        -------
        InferenceData
            ``InferenceData`` object with the groups ``prior``, ``prior_predictive`` and
            ``observed_data``.
        """
        if var_names is None:
            variables = self.backend.model.unobserved_RVs + self.backend.model.observed_RVs
            variables_names = [v.name for v in variables]
            var_names = pm.util.get_default_varnames(variables_names, include_transformed=False)

        if omit_offsets:
            var_names = [name for name in var_names if not name.endswith("_offset")]

        idata = pm.sample_prior_predictive(
            samples=draws, var_names=var_names, model=self.backend.model, random_seed=random_seed
        )

        if hasattr(idata, "prior"):
            to_drop = [dim for dim in idata.prior.dims if dim.endswith("_dim_0")]
            idata.prior = idata.prior.squeeze(to_drop).reset_coords(to_drop, drop=True)

        for group in idata.groups():
            getattr(idata, group).attrs["modeling_interface"] = "bambi"
            getattr(idata, group).attrs["modeling_interface_version"] = __version__

        return idata

    # pylint: disable=protected-access
    def predict(self, idata, kind="mean", data=None, inplace=True, include_group_specific=True):
        """Predict method for Bambi models

        Obtains in-sample and out-of-sample predictions from a fitted Bambi model.

        Parameters
        ----------
        idata: InferenceData
            The ``InferenceData`` instance returned by ``.fit()``.
        kind: str
            Indicates the type of prediction required. Can be ``"mean"`` or ``"pps"``. The
            first returns draws from the posterior distribution of the mean, while the latter
            returns the draws from the posterior predictive distribution
            (i.e. the posterior probability distribution for a new observation).
            Defaults to ``"mean"``.
        data: pandas.DataFrame or None
            An optional data frame with values for the predictors that are used to obtain
            out-of-sample predictions. If omitted, the original dataset is used.
        include_group_specific: bool
            If ``True`` make predictions including the group specific effects. Otherwise,
            predictions are made with common effects only (i.e. group specific are set
            to zero).
        inplace: bool
            If ``True`` it will modify ``idata`` in-place. Otherwise, it will return a copy of
            ``idata`` with the predictions added. If ``kind="mean"``, a new variable ending in
            ``"_mean"`` is added to the ``posterior`` group. If ``kind="pps"``, it appends a
            ``posterior_predictive`` group to ``idata``. If any of these already exist, it will be
            overwritten.

        Returns
        -------
        InferenceData or None
        """

        if kind not in ["mean", "pps"]:
            raise ValueError("'kind' must be one of 'mean' or 'pps'")

        linear_predictor = 0
        X = None
        Z = None
        x_offsets = []

        chain_n = len(idata.posterior.coords.get("chain"))
        draw_n = len(idata.posterior.coords.get("draw"))
        posterior = idata.posterior
        in_sample = data is None

        if not inplace:
            idata = deepcopy(idata)

        # Create design matrices
        if self._design.common:
            if in_sample:
                X = self._design.common.design_matrix
            else:
                X = self._design.common.evaluate_new_data(data).design_matrix

            # Add offset columns to their own design matrix
            # Remove them from the common design matrix.
            for term in self.offset_terms:
                term_slice = self._design.common.slices[term]
                x_offsets.append(X[:, term_slice])
                X = np.delete(X, term_slice, axis=1)

        if self._design.group and include_group_specific:
            if in_sample:
                Z = self._design.group.design_matrix
            else:
                Z = self._design.group.evaluate_new_data(data).design_matrix

        # Contribution due to common terms
        if X is not None:
            beta_x_list = []
            term_names = list(self.common_terms)
            response_dims = list(self.response.coords)

            if self.intercept_term:
                term_names.insert(0, "Intercept")

            for name in term_names:
                term_dims = list(self.terms[name].coords)
                term_posterior = posterior[name]
                dims = set(term_posterior.coords)

                # 1-dimensional predictors (a single slope or intercept)
                if dims == {"chain", "draw"}:
                    values = term_posterior.stack(samples=("chain", "draw")).values
                    if len(values.shape) == 1:
                        values = values[:, np.newaxis]
                # 2-dimensional predictors (splines or categoricals)
                elif dims == {"chain", "draw"}.union(term_dims):
                    transpose_dims = ["samples"] + term_dims
                    values = (
                        term_posterior.stack(samples=("chain", "draw"))
                        .transpose(*transpose_dims)
                        .values
                    )
                    if len(values.shape) == 1:
                        values = values[:, np.newaxis]
                # Multivariate families, we need to consider the dimension of the response too
                elif isinstance(self.family, (multivariate.Categorical, multivariate.Multinomial)):
                    transpose_dims = ["samples"] + term_dims + response_dims
                    values = (
                        term_posterior.stack(samples=("chain", "draw"))
                        .transpose(*transpose_dims)
                        .values
                    )
                    # When p = 1 values is of shape (samples_n, response_n).
                    # We need it to be of shape (samples_n, 1, response_n)
                    if len(values.shape) == 2:
                        values = values[:, np.newaxis, :]
                else:
                    raise ValueError(f"Unexpected dimensions in term {name}")

                beta_x_list.append(values)

            # 'beta_x' is of shape:
            # * (chain_n * draw_n, p) for univariate
            # * (chain_n * draw_n, p, response_n) for multivariate
            beta_x = np.hstack(beta_x_list)

            # 'contribution' is of shape:
            # * (chain_n * draw_n, obs_n) for univariate
            # * (chain_n * draw_n, obs_n, response_n) for multivariate
            if len(beta_x.shape) == 2:
                contribution = np.dot(X, beta_x.T).T
            else:
                contribution = np.zeros((beta_x.shape[0], X.shape[0], beta_x.shape[2]))
                for i in range(contribution.shape[2]):
                    contribution[:, :, i] = np.dot(X, beta_x[:, :, i].T).T

            shape = (chain_n, draw_n) + contribution.shape[1:]
            contribution = contribution.reshape(shape)
            linear_predictor += contribution

        # If model contains offset, add directly to the linear predictor
        if x_offsets:
            linear_predictor += np.column_stack(x_offsets).sum(axis=1)[np.newaxis, np.newaxis, :]

        # Contribution due to group-specific terms. Same comments than for beta_x apply here.
        if Z is not None:
            beta_z_list = []
            term_names = list(self.group_specific_terms)
            response_dims = list(self.response.coords)

            for name in term_names:
                term_dims = list(self.terms[name].coords)
                factor_dims = [c for c in term_dims if c.endswith("__factor_dim")]
                expr_dims = [c for c in term_dims if c.endswith("__expr_dim")]
                term_posterior = posterior[name]
                dims = set(term_posterior.dims)

                # Group-specific term: len(dims) < 3 does not exist.
                # 1 dimensional predictors
                if dims == {"chain", "draw"}.union(expr_dims):
                    transpose_dims = ["samples"] + expr_dims
                    values = (
                        term_posterior.stack(samples=("chain", "draw"))
                        .transpose(*transpose_dims)
                        .values
                    )
                # 2 dimensional predictors
                elif dims == {"chain", "draw"}.union(expr_dims + factor_dims):
                    transpose_dims = ["samples", "coefs"]
                    values = (
                        term_posterior.stack(samples=("chain", "draw"))
                        .stack(coefs=tuple(factor_dims + expr_dims))
                        .transpose(*transpose_dims)
                        .values
                    )
                # Multivariate families, need to consider dimensionality of the response
                elif isinstance(self.family, (multivariate.Categorical, multivariate.Multinomial)):
                    # 1 dimensional predictors (there's no factor dimension)
                    if dims == {"chain", "draw"}.union(factor_dims + response_dims):
                        transpose_dims = ["samples"] + factor_dims + response_dims
                        values = (
                            term_posterior.stack(samples=("chain", "draw"))
                            .transpose(*transpose_dims)
                            .values
                        )
                    # 2 dimensional predictors (there's a factor dimension)
                    elif dims == {"chain", "draw"}.union(expr_dims + factor_dims + response_dims):
                        transpose_dims = ["samples", "coefs"] + response_dims
                        values = (
                            term_posterior.stack(samples=("chain", "draw"))
                            .stack(coefs=tuple(factor_dims + expr_dims))
                            .transpose(*transpose_dims)
                            .values
                        )
                    else:
                        raise ValueError(f"Unexpected dimensions in term {name}")
                else:
                    raise ValueError(f"Unexpected dimensions in term {name}")

                beta_z_list.append(values)

            # 'beta_z' is of shape:
            # * (chain_n * draw_n, p) for univariate
            # * (chain_n * draw_n, p, response_n) for multivariate models
            beta_z = np.hstack(beta_z_list)

            # 'contribution' is of shape:
            # * (chain_n * draw_n, obs_n) for univariate
            # * (chain_n * draw_n, obs_n, response_n) for multivariate
            if len(beta_z.shape) == 2:
                contribution = np.dot(Z, beta_z.T).T
            else:
                contribution = np.zeros((beta_z.shape[0], Z.shape[0], beta_z.shape[2]))
                for i in range(contribution.shape[2]):
                    contribution[:, :, i] = np.dot(Z, beta_z[:, :, i].T).T

            shape = (chain_n, draw_n) + contribution.shape[1:]
            contribution = contribution.reshape(shape)
            linear_predictor += contribution

        if kind == "mean":
            idata.posterior = self.family.predict(self, posterior, linear_predictor)
        else:
            pps_kwargs = {
                "model": self,
                "posterior": posterior,
                "linear_predictor": linear_predictor,
            }

            if not in_sample and isinstance(self.family, univariate.Binomial):
                pps_kwargs["trials"] = self._design.response.evaluate_new_data(data)

            pps = self.family.posterior_predictive(**pps_kwargs)

            if "posterior_predictive" in idata:
                del idata.posterior_predictive
            idata.add_groups({"posterior_predictive": {self.response.name: pps}})
            getattr(idata, "posterior_predictive").attrs["modeling_interface"] = "bambi"
            getattr(idata, "posterior_predictive").attrs["modeling_interface_version"] = __version__

        if inplace:
            return None
        else:
            return idata

    def graph(self, formatting="plain", name=None, figsize=None, dpi=300, fmt="png"):
        """
        Produce a graphviz Digraph from a built Bambi model.

        Requires graphviz, which may be installed most easily with
            ``conda install -c conda-forge python-graphviz``

        Alternatively, you may install the ``graphviz`` binaries yourself, and then
        ``pip install graphviz`` to get the python bindings.
        See http://graphviz.readthedocs.io/en/stable/manual.html for more information.

        Parameters
        ----------
        formatting: str
            One of ``"plain"`` or ``"plain_with_params"``. Defaults to ``"plain"``.
        name: str
            Name of the figure to save. Defaults to ``None``, no figure is saved.
        figsize: tuple
            Maximum width and height of figure in inches. Defaults to ``None``, the figure size is
            set automatically. If defined and the drawing is larger than the given size, the drawing
            is uniformly scaled down so that it fits within the given size.  Only works if ``name``
            is not ``None``.
        dpi: int
            Point per inch of the figure to save.
            Defaults to 300. Only works if ``name`` is not ``None``.
        fmt: str
            Format of the figure to save.
            Defaults to ``"png"``. Only works if ``name`` is not ``None``.

        Example
        --------
        >>> model = Model("y ~ x + (1|z)")
        >>> model.build()
        >>> model.graph()

        >>> model = Model("y ~ x + (1|z)")
        >>> model.fit()
        >>> model.graph()

        """
        if self.backend is None:
            raise ValueError(
                "The model is empty. "
                "Are you forgetting to first call .build() or .fit() on the Bambi model?"
            )

        graphviz = pm.model_to_graphviz(model=self.backend.model, formatting=formatting)

        width, height = (None, None) if figsize is None else figsize

        if name is not None:
            graphviz_ = graphviz.copy()
            graphviz_.graph_attr.update(size=f"{width},{height}!")
            graphviz_.graph_attr.update(dpi=str(dpi))
            graphviz_.render(filename=name, format=fmt, cleanup=True)

        return graphviz

    def _get_group_specific_groups(self):
        groups = {}
        for term_name in self.group_specific_terms:
            factor = term_name.split("|")[1]
            if factor not in groups:
                groups[factor] = [term_name]
            else:
                groups[factor].append(term_name)
        return groups

    def __str__(self):
        priors_common = [f"    {t.name} ~ {t.prior}" for t in self.common_terms.values()]
        if self.intercept_term:
            term = self.intercept_term
            priors_common = [f"    {term.name} ~ {term.prior}"] + priors_common

        priors_group = [f"    {t.name} ~ {t.prior}" for t in self.group_specific_terms.values()]

        # Prior for the correlation matrix in group-specific terms
        priors_cor = [f"    {k} ~ LKJCorr({v})" for k, v in self.priors_cor.items()]

        # Priors for auxiliary parameters, e.g., standard deviation in normal linear model
        priors_aux = [f"    {k} ~ {v}" for k, v in self.family.likelihood.priors.items()]

        # Offsets
        offsets = [f"    {t.name} ~ 1" for t in self.offset_terms.values()]

        priors_dict = {
            "Common-level effects": priors_common,
            "Group-level effects": priors_group,
            "Group-level correlation": priors_cor,
            "Offset effects": offsets,
            "Auxiliary parameters": priors_aux,
        }

        priors_list = []
        for group, values in priors_dict.items():
            if values:
                priors_list += ["\n".join([f"  {group}"] + values)]
        priors_message = "\n\n".join(priors_list)

        str_list = [
            f"Formula: {self.formula}",
            f"Family name: {self.family.name.capitalize()}",
            f"Link: {self.family.link.name}",
            f"Observations: {self.response.data.shape[0]}",
            "Priors:",
            priors_message,
        ]
        if self.backend and self.backend.fit:
            extra_foot = (
                "------\n"
                "* To see a plot of the priors call the .plot_priors() method.\n"
                "* To see a summary or plot of the posterior pass the object returned "
                "by .fit() to az.summary() or az.plot_trace()\n"
            )
            str_list += [extra_foot]

        return "\n".join(str_list)

    def __repr__(self):
        return self.__str__()

    @property
    def term_names(self):
        """Return names of all terms in order of addition to model."""
        return list(self.terms)

    @property
    def common_terms(self):
        """Return dict of all common effects in model."""
        return {
            k: v
            for (k, v) in self.terms.items()
            if not v.group_specific and v.kind not in ["intercept", "offset"]
        }

    @property
    def group_specific_terms(self):
        """Return dict of all group specific effects in model."""
        return {k: v for (k, v) in self.terms.items() if v.group_specific}

    @property
    def intercept_term(self):
        """Return the intercept term"""
        term = [v for v in self.terms.values() if not v.group_specific and v.kind == "intercept"]
        if term:
            return term[0]
        else:
            return None

    @property
    def offset_terms(self):
        """Return dict of all offset effects in model."""
        return {
            k: v for (k, v) in self.terms.items() if not v.group_specific and v.kind == "offset"
        }


def prepare_prior(prior, kind, auto_scale):
    """Helper function to correctly set default priors, auto scaling, etc.

    Parameters
    ----------
    prior : Prior, float, or None.
    kind : string
        Accepted values are: ``"intercept"``, ``"common"``, or ``"group_specific"``.
    """
    if prior is None and not auto_scale:
        prior = get_default_prior(kind + "_flat")
    if isinstance(prior, Prior):
        prior.auto_scale = False
    else:
        scale = prior
        prior = get_default_prior(kind)
        prior.scale = scale
    return prior



import numpy as np
import aesara.tensor as at


def probit(x):
    """Probit function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 0.5 + 0.5 * at.erf(x / at.sqrt(2))
    result = at.switch(at.eq(result, 0), eps, result)
    result = at.switch(at.eq(result, 1), 1 - eps, result)

    return result


def cloglog(x):
    """Cloglog function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 1 - at.exp(-at.exp(x))
    result = at.switch(at.eq(result, 0), eps, result)
    result = at.switch(at.eq(result, 1), 1 - eps, result)

    return result


def logit(x):
    """Logit function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = at.sigmoid(x)
    result = at.switch(at.eq(result, 0), eps, result)
    result = at.switch(at.eq(result, 1), 1 - eps, result)
    return result


def identity(x):
    return x


def inverse_squared(x):
    return at.reciprocal(at.sqrt(x))


def arctan_2(x):
    return 2 * at.arctan(x)



import numpy as np

from bambi.families.univariate import Gaussian, StudentT, VonMises

from .prior import Prior


class PriorScaler:
    """Scale prior distributions parameters."""

    # Standard deviation multiplier.
    STD = 2.5

    def __init__(self, model):
        self.model = model
        self.has_intercept = model.intercept_term is not None
        self.priors = {}

        # Compute mean and std of the response
        if isinstance(self.model.family, (Gaussian, StudentT)):
            self.response_mean = np.mean(model.response.data)
            self.response_std = np.std(self.model.response.data)
        else:
            self.response_mean = 0
            self.response_std = 1

    def get_intercept_stats(self):
        mu = self.response_mean
        sigma = self.STD * self.response_std

        # Only adjust mu and sigma if there is at least one Normal prior for a common term.
        if self.priors:
            sigmas = np.hstack([prior["sigma"] for prior in self.priors.values()])
            x_mean = np.hstack([self.model.terms[term].data.mean(axis=0) for term in self.priors])
            sigma = (sigma**2 + np.dot(sigmas**2, x_mean**2)) ** 0.5

        return mu, sigma

    def get_slope_sigma(self, x):
        return self.STD * (self.response_std / np.std(x))

    def scale_response(self):
        # Add cases for other families
        priors = self.model.family.likelihood.priors
        if isinstance(self.model.family, (Gaussian, StudentT)):
            if priors["sigma"].auto_scale:
                priors["sigma"] = Prior("HalfStudentT", nu=4, sigma=self.response_std)
        elif isinstance(self.model.family, VonMises):
            if priors["kappa"].auto_scale:
                priors["kappa"] = Prior("HalfStudentT", nu=4, sigma=self.response_std)

    def scale_intercept(self, term):
        if term.prior.name != "Normal":
            return
        mu, sigma = self.get_intercept_stats()
        term.prior.update(mu=mu, sigma=sigma)

    def scale_common(self, term):
        if term.prior.name != "Normal":
            return

        # As many zeros as columns in the data. It can be greater than 1 for categorical variables
        mu = np.zeros(term.data.shape[1])
        sigma = np.zeros(term.data.shape[1])

        # Iterate over columns in the data
        for i, x in enumerate(term.data.T):
            sigma[i] = self.get_slope_sigma(x)

        # Save and set prior
        self.priors.update({term.name: {"mu": mu, "sigma": sigma}})
        term.prior.update(mu=mu, sigma=sigma)

    def scale_group_specific(self, term):
        if term.prior.args["sigma"].name != "HalfNormal":
            return

        # Handle intercepts
        if term.kind == "intercept":
            _, sigma = self.get_intercept_stats()
        # Handle slopes
        else:
            # Recreate the corresponding common effect data
            if len(term.predictor.shape) == 2:
                data_as_common = term.predictor
            else:
                data_as_common = term.predictor[:, None]
            sigma = np.zeros(data_as_common.shape[1])
            for i, x in enumerate(data_as_common.T):
                sigma[i] = self.get_slope_sigma(x)
        term.prior.args["sigma"].update(sigma=np.squeeze(np.atleast_1d(sigma)))

    def scale(self):

        # Scale response
        self.scale_response()

        # Scale common terms
        for term in self.model.common_terms.values():
            if term.prior.auto_scale:
                self.scale_common(term)

        # Scale intercept
        if self.has_intercept:
            term = self.model.intercept_term
            if term.prior.auto_scale:
                self.scale_intercept(term)

        # Scale group-specific terms
        for term in self.model.group_specific_terms.values():
            if term.prior.auto_scale:
                self.scale_group_specific(term)



from collections import namedtuple

import numpy as np

from scipy import special

from bambi.utils import multilinify, spacify


def force_within_unit_interval(x):
    """Make sure data in unit interval is in (0, 1)"""
    eps = np.finfo(float).eps
    x[x == 0] = eps
    x[x == 1] = 1 - eps
    return x


def force_greater_than_zero(x):
    """Make sure data in positive reals is in (0, infty)"""
    eps = np.finfo(float).eps
    x[x == 0] = eps
    return x


def identity(x):
    return x


def cloglog(mu):
    """Cloglog function that ensures the input is greater than 0."""
    mu = force_greater_than_zero(mu)
    return np.log(-np.log(1 - mu))


def invcloglog(eta):
    """Inverse of the cloglog function that ensures result is in (0, 1)"""
    result = 1 - np.exp(-np.exp(eta))
    return force_within_unit_interval(result)


def probit(mu):
    """Probit function that ensures the input is in (0, 1)"""
    mu = force_within_unit_interval(mu)
    return 2**0.5 * special.erfinv(2 * mu - 1)  # pylint: disable=no-member


def invprobit(eta):
    """Inverse of the probit function that ensures result is in (0, 1)"""
    result = 0.5 + 0.5 * special.erf(eta / 2**0.5)  # pylint: disable=no-member
    return force_within_unit_interval(result)


def expit(eta):
    """Expit function that ensures result is in (0, 1)"""
    result = special.expit(eta)  # pylint: disable=no-member
    result = force_within_unit_interval(result)
    return result


def logit(mu):
    """Logit function that ensures the input is in (0, 1)"""
    mu = force_within_unit_interval(mu)
    return special.logit(mu)  # pylint: disable=no-member


def softmax(eta, axis=None):
    result = special.softmax(eta, axis=axis)  # pylint: disable=no-member
    result = force_within_unit_interval(result)
    return result


def inverse_squared(mu):
    return 1 / mu**2


def inv_inverse_squared(eta):
    return 1 / np.sqrt(eta)


def arctan_2(eta):
    return 2 * np.arctan(eta)


def tan_2(mu):
    return np.tan(mu / 2)


def inverse(mu):
    return 1 / mu


def inv_inverse(eta):
    return 1 / eta


def link_not_implemented(*args, **kwargs):
    raise NotImplementedError("link not implemented")


# link: Known as g in the GLM literature. Maps the response to the linear predictor scale.
# linkinv: Known as g^(-1) in the GLM literature. Maps the linear predictor to the response scale.
LinksContainer = namedtuple("LinksContainer", ["link", "linkinv"])

LINKS = {
    "cloglog": LinksContainer(cloglog, invcloglog),
    "identity": LinksContainer(identity, identity),
    "inverse_squared": LinksContainer(inverse_squared, inv_inverse_squared),
    "inverse": LinksContainer(inverse, inv_inverse),
    "log": LinksContainer(np.log, np.exp),
    "logit": LinksContainer(logit, expit),
    "probit": LinksContainer(probit, invprobit),
    "softmax": LinksContainer(link_not_implemented, softmax),
    "tan_2": LinksContainer(tan_2, arctan_2),
}


class Link:
    """Representation of a link function.

    This object contains two main functions. One is the link function itself, the function
    that maps values in the response scale to the linear predictor, and the other is the inverse
    of the link function, that maps values of the linear predictor to the response scale.

    The great majority of users will never interact with this class unless they want to create
    a custom ``Family`` with a custom ``Link``. This is automatically handled for all the built-in
    families.

    Parameters
    ----------
    name: str
        The name of the link function. If it is a known name, it's not necessary to pass any
        other arguments because functions are already defined internally. If not known, all of
        ``link``, ``linkinv`` and ``linkinv_backend`` must be specified.
    link: function
        A function that maps the response to the linear predictor. Known as the :math:`g` function
        in GLM jargon. Does not need to be specified when ``name`` is a known name.
    linkinv: function
        A function that maps the linear predictor to the response. Known as the :math:`g^{-1}`
        function in GLM jargon. Does not need to be specified when ``name`` is a known name.
    linkinv_backend: function
        Same than ``linkinv`` but must be something that works with PyMC backend (i.e. it must
        work with Aesara tensors). Does not need to be specified when ``name`` is a known
        name.
    """

    def __init__(self, name, link=None, linkinv=None, linkinv_backend=None):
        self.name = name
        self.link = link
        self.linkinv = linkinv
        self.linkinv_backend = linkinv_backend

        if name in LINKS:
            self.link = LINKS[name].link
            self.linkinv = LINKS[name].linkinv
        else:
            if not link or not linkinv or not linkinv_backend:
                raise ValueError(
                    f"Link name '{name}' is not supported and at least one of 'link', "
                    "'linkinv' or 'linkinv_backend' are unspecified."
                )

    def __str__(self):
        args = [f"name: {self.name}", f"link: {self.link}", f"linkinv: {self.linkinv}"]
        return f"{self.__class__.__name__}({spacify(multilinify(args))}\n)"

    def __repr__(self):
        return self.__str__()



# pylint: disable=unused-argument
import numpy as np

from .family import Family


class MultivariateFamily(Family):
    def predict(self, model, posterior, linear_predictor):
        return NotImplemented

    def posterior_predictive(self, model, posterior, linear_predictor):
        return NotImplemented


class Categorical(MultivariateFamily):
    SUPPORTED_LINKS = ["softmax"]

    def predict(self, model, posterior, linear_predictor):
        # This is only 'softmax' for now.
        # Last axis is the one where the response coord is inserted
        # We need to append zeros for the reference category
        shape = linear_predictor.shape
        linear_predictor = np.concatenate([np.zeros(shape[:-1] + (1,)), linear_predictor], axis=-1)

        mean = self.link.linkinv(linear_predictor, axis=-1)

        obs_n = mean.shape[2]
        response_var = model.response.name + "_mean"
        response_dim = model.response.name + "_obs"

        # Drop var/dim if already present
        if response_var in posterior.data_vars:
            posterior = posterior.drop_vars(response_var)

        if response_dim in posterior.dims:
            posterior = posterior.drop_dims(response_dim)

        response_levels_dim = model.response.name + "_mean_dim"
        dims = ("chain", "draw", response_dim, response_levels_dim)
        posterior[response_var] = (dims, mean)

        posterior = posterior.assign_coords({response_levels_dim: model.response.levels})
        posterior = posterior.assign_coords({response_dim: list(range(obs_n))})
        return posterior

    def posterior_predictive(self, model, posterior, linear_predictor):
        # https://stackoverflow.com/questions/34187130
        def draw_categorical_samples(probability_matrix, items):
            cumsum = probability_matrix.cumsum(axis=0)
            idx = np.random.rand(probability_matrix.shape[1])
            idx = (cumsum < idx).sum(axis=0)
            return items[idx]

        shape = linear_predictor.shape
        linear_predictor = np.concatenate([np.zeros(shape[:-1] + (1,)), linear_predictor], axis=-1)

        mean = self.link.linkinv(linear_predictor, axis=-1)
        shape = mean.shape

        mean = mean.reshape((mean.shape[0] * mean.shape[1], mean.shape[2], mean.shape[3]))
        draws_n = mean.shape[0]
        obs_n = mean.shape[1]

        pps = np.empty((draws_n, obs_n), dtype=int)
        response_levels = np.arange(len(model.response.levels))
        for idx in range(obs_n):
            pps[:, idx] = draw_categorical_samples(mean[:, idx, :].T, response_levels)

        return pps.reshape((shape[0], shape[1], obs_n))


class Multinomial(MultivariateFamily):
    SUPPORTED_LINKS = ["softmax"]

    def predict(self, model, posterior, linear_predictor):
        # This is only 'softmax' for now.
        # Last axis is the one where the response coord is inserted
        # We need to append zeros for the reference category
        shape = linear_predictor.shape
        linear_predictor = np.concatenate([np.zeros(shape[:-1] + (1,)), linear_predictor], axis=-1)

        mean = self.link.linkinv(linear_predictor, axis=-1)

        obs_n = mean.shape[2]
        response_var = model.response.name + "_mean"
        response_dim = model.response.name + "_obs"

        # Drop var/dim if already present
        if response_var in posterior.data_vars:
            posterior = posterior.drop_vars(response_var)

        if response_dim in posterior.dims:
            posterior = posterior.drop_dims(response_dim)

        response_levels_dim = model.response.name + "_mean_dim"
        dims = ("chain", "draw", response_dim, response_levels_dim)
        posterior[response_var] = (dims, mean)

        posterior = posterior.assign_coords({response_levels_dim: model.response.levels})
        posterior = posterior.assign_coords({response_dim: list(range(obs_n))})
        return posterior

    def posterior_predictive(self, model, posterior, linear_predictor):
        shape = linear_predictor.shape
        linear_predictor = np.concatenate([np.zeros(shape[:-1] + (1,)), linear_predictor], axis=-1)

        mean = self.link.linkinv(linear_predictor, axis=-1)
        shape = mean.shape

        mean = mean.reshape((mean.shape[0] * mean.shape[1], mean.shape[2], mean.shape[3]))
        draws_n = mean.shape[0]
        obs_n = mean.shape[1]

        pps = np.empty(mean.shape, dtype=int)
        n = model.response.data.sum(1)

        # random.multinomial only accepts
        # * n : integer
        # * p : vector
        for i in range(obs_n):
            for j in range(draws_n):
                pps[j, i, :] = np.random.multinomial(n[i], mean[j, i, :])

        # Final shape is of (chain, draw, obs_n, response_n)
        pps = pps.reshape(shape)
        return pps



from collections import namedtuple

from bambi.priors import Prior
from bambi.utils import multilinify, spacify


DistSettings = namedtuple("DistSettings", ["params", "parent", "args"])

DISTRIBUTIONS = {
    "Bernoulli": DistSettings(params=("p",), parent="p", args=None),
    "Beta": DistSettings(params=("mu", "kappa"), parent="mu", args=("kappa",)),
    "Binomial": DistSettings(params=("p",), parent="p", args=None),
    "Categorical": DistSettings(params=("p",), parent="p", args=None),
    "Gamma": DistSettings(params=("mu", "alpha"), parent="mu", args=("alpha",)),
    "Multinomial": DistSettings(params=("p",), parent="p", args=None),
    "Normal": DistSettings(params=("mu", "sigma"), parent="mu", args=("sigma",)),
    "NegativeBinomial": DistSettings(params=("mu", "alpha"), parent="mu", args=("alpha",)),
    "Laplace": DistSettings(params=("mu", "b"), parent="mu", args=("b",)),
    "Poisson": DistSettings(params=("mu",), parent="mu", args=None),
    "StudentT": DistSettings(params=("mu", "sigma"), parent="mu", args=("sigma", "nu")),
    "VonMises": DistSettings(params=("mu", "kappa"), parent="mu", args=("kappa",)),
    "Wald": DistSettings(params=("mu", "lam"), parent="mu", args=("lam",)),
}


class Likelihood:
    """Representation of a Likelihood function for a Bambi model.

    Notes:
    * ``parent`` must not be in ``kwargs``.
    * ``parent`` is inferred from the ``name`` if it is a known name

    Parameters
    ----------
    name: str
        Name of the likelihood function. Must be a valid PyMC distribution name.
    parent: str
        Optional specification of the name of the mean parameter in the likelihood.
        This is the parameter whose transformation is modeled by the linear predictor.
    kwargs:
        Keyword arguments that indicate prior distributions for auxiliary parameters in the
        likelihood.
    """

    DISTRIBUTIONS = DISTRIBUTIONS

    def __init__(self, name, parent=None, **kwargs):
        if name in self.DISTRIBUTIONS:
            self.name = name
            self.parent = parent
            self.priors = self._check_priors(kwargs)
        else:
            # On your own risk
            self.name = name
            # Check priors passed are in fact of class Prior
            check_all_are_priors(kwargs)
            self.priors = kwargs
            self.parent = parent

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, x):
        # Checks are made when using a known distribution
        if self.name in self.DISTRIBUTIONS:
            if x is None:
                x = self.DISTRIBUTIONS[self.name].parent
            elif x not in self.DISTRIBUTIONS[self.name].params:
                raise ValueError(f"'{x}' is not a valid parameter for the likelihood '{self.name}'")
        # Otherwise, no check is done. At your own risk!
        self._parent = x

    def _check_priors(self, priors):
        args = self.DISTRIBUTIONS[self.name].args

        # The function requires priors but none were passed
        if priors == {} and args is not None:
            raise ValueError(f"'{self.name}' requires priors for the parameters {args}.")

        # The function does not require priors, but at least one was passed
        if priors != {} and args is None:
            raise ValueError(f"'{self.name}' does not require any additional priors.")

        # The function requires priors, priors were passed, but they differ from the required
        if priors and args:
            difference = set(args) - set(priors)
            if len(difference):
                raise ValueError(f"'{self.name}' misses priors for the parameters {difference}")

            # And check priors passed are in fact of class Prior
            check_all_are_priors(priors)

        return priors

    def __str__(self):
        args = [f"name: {self.name}", f"parent: {self.parent}", f"priors: {self.priors}"]
        return f"{self.__class__.__name__}({spacify(multilinify(args))}\n)"

    def __repr__(self):
        return self.__str__()


def check_all_are_priors(priors):
    """Checks if values in the supplied dictionary are all valid prior objects

    An object is a valid prior if
    * It is an instance of bambi.priors.Prior
    * It is a number

    Parameters
    ----------
    priors: dict
        A dictionary whose values are tested to be valid priors
    """
    if any(not isinstance(prior, (Prior, int, float)) for prior in priors.values()):
        raise ValueError("Prior distributions must be a 'Prior' instance or a numeric value")



import aesara.tensor as at
import pymc as pm


def get_distribution(dist):
    """Return a PyMC distribution."""
    if isinstance(dist, str):
        if hasattr(pm, dist):
            dist = getattr(pm, dist)
        else:
            raise ValueError(f"The Distribution '{dist}' was not found in PyMC")
    return dist


def has_hyperprior(kwargs):
    """Determines if a Prior has an hyperprior"""
    return (
        "sigma" in kwargs
        and "observed" not in kwargs
        and isinstance(kwargs["sigma"], at.TensorVariable)
    )



import logging

from pymc import math

from .data import clear_data_home, load_data
from .families import Family, Likelihood, Link
from .models import Model
from .priors import Prior
from .backend import PyMCModel
from .version import __version__


__all__ = ["Model", "Prior", "Family", "Likelihood", "Link", "PyMCModel"]

_log = logging.getLogger("bambi")

if not logging.root.handlers:
    _log.setLevel(logging.INFO)
    if len(_log.handlers) == 0:
        handler = logging.StreamHandler()
        _log.addHandler(handler)



import logging
import traceback
from copy import deepcopy

import numpy as np
import pymc as pm

import aesara.tensor as at

from bambi import version

from bambi.backend.links import cloglog, identity, inverse_squared, logit, probit, arctan_2
from bambi.backend.terms import CommonTerm, GroupSpecificTerm, InterceptTerm, ResponseTerm
from bambi.families.multivariate import Categorical, Multinomial

_log = logging.getLogger("bambi")


class PyMCModel:
    """PyMC model-fitting backend."""

    INVLINKS = {
        "cloglog": cloglog,
        "identity": identity,
        "inverse_squared": inverse_squared,
        "inverse": at.reciprocal,
        "log": at.exp,
        "logit": logit,
        "probit": probit,
        "tan_2": arctan_2,
        "softmax": at.nnet.softmax,
    }

    def __init__(self):
        self.name = pm.__name__
        self.version = pm.__version__

        # Attributes defined elsewhere
        self._design_matrix_without_intercept = None
        self.vi_approx = None
        self.coords = {}
        self.fit = False
        self.has_intercept = False
        self.model = None
        self.mu = None
        self.spec = None

    def build(self, spec):
        """Compile the PyMC model from an abstract model specification.

        Parameters
        ----------
        spec: bambi.Model
            A Bambi ``Model`` instance containing the abstract specification of the model
            to compile.
        """
        self.model = pm.Model()
        self.has_intercept = spec.intercept_term is not None
        self.mu = 0.0

        for name, values in spec.response.coords.items():
            if name not in self.model.coords:
                self.model.add_coords({name: values})
        self.coords.update(**spec.response.coords)

        with self.model:
            self._build_intercept(spec)
            self._build_offsets(spec)
            self._build_common_terms(spec)
            self._build_group_specific_terms(spec)
            self._build_response(spec)
            self._build_potentials(spec)
        self.spec = spec

    def run(
        self,
        draws=1000,
        tune=1000,
        discard_tuned_samples=True,
        omit_offsets=True,
        include_mean=False,
        inference_method="mcmc",
        init="auto",
        n_init=50000,
        chains=None,
        cores=None,
        random_seed=None,
        **kwargs,
    ):
        """Run PyMC sampler."""
        inference_method = inference_method.lower()
        # NOTE: Methods return different types of objects (idata, approximation, and dictionary)
        if inference_method in ["mcmc", "nuts_numpyro", "nuts_blackjax"]:
            result = self._run_mcmc(
                draws,
                tune,
                discard_tuned_samples,
                omit_offsets,
                include_mean,
                init,
                n_init,
                chains,
                cores,
                random_seed,
                inference_method,
                **kwargs,
            )
        elif inference_method == "vi":
            result = self._run_vi(**kwargs)
        elif inference_method == "laplace":
            result = self._run_laplace(draws, omit_offsets, include_mean)
        else:
            raise NotImplementedError(f"{inference_method} method has not been implemented")

        self.fit = True
        return result

    def _build_intercept(self, spec):
        """Add intercept term to the PyMC model.

        We have linear predictors of the form 'X @ b + Z @ u'. This is technically part of
        'X @ b' but it is added separately for convenience reasons.

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        if self.has_intercept:
            self.mu += InterceptTerm(spec.intercept_term).build(spec)

    def _build_common_terms(self, spec):
        """Add common (fixed) terms to the PyMC model.

        We have linear predictors of the form 'X @ b + Z @ u'.
        This creates the 'b' parameter vector in PyMC, computes `X @ b`, and adds it to ``self.mu``.

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        if spec.common_terms:
            coefs = []
            columns = []
            for term in spec.common_terms.values():
                common_term = CommonTerm(term)
                # Add coords
                for name, values in common_term.coords.items():
                    if name not in self.model.coords:
                        self.model.add_coords({name: values})
                self.coords.update(**common_term.coords)

                # Build
                coef, data = common_term.build(spec)
                coefs.append(coef)
                columns.append(data)

            # Column vector of coefficients and design matrix
            coefs = at.concatenate(coefs)
            data = np.hstack(columns)

            # If there's an intercept, center the data
            # Also store the design matrix without the intercept to uncenter the intercept later
            if self.has_intercept:
                self._design_matrix_without_intercept = data
                data = data - data.mean(0)

            # Add term to linear predictor
            self.mu += at.dot(data, coefs)

    def _build_group_specific_terms(self, spec):
        """Add group-specific (random or varying) terms to the PyMC model.

        We have linear predictors of the form 'X @ b + Z @ u'.
        This creates the 'u' parameter vector in PyMC, computes `Z @ u`, and adds it to ``self.mu``.

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        # Add group specific terms that have prior for their correlation matrix
        for group, eta in spec.priors_cor.items():
            # pylint: disable=protected-access
            terms = [spec.terms[name] for name in spec._get_group_specific_groups()[group]]
            self.mu += add_lkj(self, terms, eta)

        terms = [
            term
            for term in spec.group_specific_terms.values()
            if term.name.split("|")[1] not in spec.priors_cor
        ]
        for term in terms:
            group_specific_term = GroupSpecificTerm(term, spec.noncentered)

            # Add coords
            for name, values in group_specific_term.coords.items():
                if name not in self.model.coords:
                    self.model.add_coords({name: values})
            self.coords.update(**group_specific_term.coords)

            # Build
            coef, predictor = group_specific_term.build(spec)

            # Add to the linear predictor
            # The loop through predictor columns is not the most beautiful alternative.
            # But it's the fastest. Doing matrix multiplication, pm.math.dot(data, coef), is slower.
            if predictor.ndim > 1:
                for col in range(predictor.shape[1]):
                    self.mu += coef[:, col] * predictor[:, col]
            elif isinstance(spec.family, (Categorical, Multinomial)):
                self.mu += coef * predictor[:, np.newaxis]
            else:
                self.mu += coef * predictor

    def _build_offsets(self, spec):
        """Add offset terms to the PyMC model.

        Offsets are terms with a regression coefficient of 1.
        This is technically part of  'X @ b' in the linear predictor 'X @ b + Z @ u'.
        It's added here so we avoid the creation of a constant variable in PyMC.

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        for offset in spec.offset_terms.values():
            self.mu += offset.data.squeeze()

    def _build_response(self, spec):
        """Add response term to the PyMC model

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        ResponseTerm(spec.response, spec.family).build(self.mu, self.INVLINKS)

    def _build_potentials(self, spec):
        """Add potentials to the PyMC model.

        Potentials are arbitrary quantities that are added to the model log likelihood.
        See 'Factor Potentials' in
        https://github.com/fonnesbeck/probabilistic_python/blob/main/pymc_intro.ipynb

        Parameters
        ----------
        spec : bambi.Model
            The model.
        """
        if spec.potentials is not None:
            count = 0
            for variable, constraint in spec.potentials:
                if isinstance(variable, (list, tuple)):
                    lambda_args = [self.model[var] for var in variable]
                    potential = constraint(*lambda_args)
                else:
                    potential = constraint(self.model[variable])
                pm.Potential(f"pot_{count}", potential)
                count += 1

    def _run_mcmc(
        self,
        draws=1000,
        tune=1000,
        discard_tuned_samples=True,
        omit_offsets=True,
        include_mean=False,
        init="auto",
        n_init=50000,
        chains=None,
        cores=None,
        random_seed=None,
        sampler_backend="mcmc",
        **kwargs,
    ):
        with self.model:
            if sampler_backend == "mcmc":
                try:
                    idata = pm.sample(
                        draws=draws,
                        tune=tune,
                        discard_tuned_samples=discard_tuned_samples,
                        init=init,
                        n_init=n_init,
                        chains=chains,
                        cores=cores,
                        random_seed=random_seed,
                        **kwargs,
                    )
                except (RuntimeError, ValueError):
                    if (
                        "ValueError: Mass matrix contains" in traceback.format_exc()
                        and init == "auto"
                    ):
                        _log.info(
                            "\nThe default initialization using init='auto' has failed, trying to "
                            "recover by switching to init='adapt_diag'",
                        )
                        idata = pm.sample(
                            draws=draws,
                            tune=tune,
                            discard_tuned_samples=discard_tuned_samples,
                            init="adapt_diag",
                            n_init=n_init,
                            chains=chains,
                            cores=cores,
                            random_seed=random_seed,
                            **kwargs,
                        )
                    else:
                        raise
            elif sampler_backend == "nuts_numpyro":
                # Lazy import to not force users to install Jax
                import pymc.sampling_jax  # pylint: disable=import-outside-toplevel

                if not chains:
                    chains = (
                        4  # sample_numpyro_nuts does not handle chains = None like pm.sample does
                    )
                idata = pymc.sampling_jax.sample_numpyro_nuts(
                    draws=draws,
                    tune=tune,
                    chains=chains,
                    random_seed=random_seed,
                    **kwargs,
                )
            elif sampler_backend == "nuts_blackjax":
                # Lazy import to not force users to install Jax
                import pymc.sampling_jax  # pylint: disable=import-outside-toplevel

                if not chains:
                    chains = (
                        4  # sample_blackjax_nuts does not handle chains = None like pm.sample does
                    )
                idata = pymc.sampling_jax.sample_blackjax_nuts(
                    draws=draws,
                    tune=tune,
                    chains=chains,
                    random_seed=random_seed,
                    **kwargs,
                )
            else:
                raise ValueError(
                    f"sampler_backend value {sampler_backend} is not valid. Please choose one of"
                    f"``mcmc``, ``nuts_numpyro`` or ``nuts_blackjax``"
                )

        idata = self._clean_results(idata, omit_offsets, include_mean)
        return idata

    def _clean_results(self, idata, omit_offsets, include_mean):
        for group in idata.groups():
            getattr(idata, group).attrs["modeling_interface"] = "bambi"
            getattr(idata, group).attrs["modeling_interface_version"] = version.__version__

        if omit_offsets:
            offset_vars = [var for var in idata.posterior.var() if var.endswith("_offset")]
            idata.posterior = idata.posterior.drop_vars(offset_vars)

        # Drop variables and dimensions associated with LKJ prior
        vars_to_drop = [var for var in idata.posterior.var() if var.startswith("_LKJ")]
        dims_to_drop = [dim for dim in idata.posterior.dims if dim.startswith("_LKJ")]

        idata.posterior = idata.posterior.drop_vars(vars_to_drop)
        idata.posterior = idata.posterior.drop_dims(dims_to_drop)

        # Drop and reorder coords
        # About coordinates ending with "_dim_0"
        # Coordinates that end with "_dim_0" are added automatically.
        # These represents unidimensional coordinates that are added for numerical variables.
        # These variables have a shape of 1 so we can concatenate the coefficients and multiply
        # the resulting vector with the design matrix.
        # But having a unidimensional coordinate for a numeric variable does not make sense.
        # So we drop them.
        coords_to_drop = [dim for dim in idata.posterior.dims if dim.endswith("_dim_0")]
        idata.posterior = idata.posterior.squeeze(coords_to_drop).reset_coords(
            coords_to_drop, drop=True
        )

        # This does not add any new coordinate, it just changes the order so the ones
        # ending in "__factor_dim" are placed after the others.
        dims_original = list(self.coords)
        dims_group = [c for c in dims_original if c.endswith("__factor_dim")]

        # Keep the original order in dims_original
        dims_original_set = set(dims_original) - set(dims_group)
        dims_original = [c for c in dims_original if c in dims_original_set]
        dims_new = ["chain", "draw"] + dims_original + dims_group
        idata.posterior = idata.posterior.transpose(*dims_new)

        # Compute the actual intercept
        if self.has_intercept and self.spec.common_terms:
            chain_n = len(idata.posterior["chain"])
            draw_n = len(idata.posterior["draw"])
            shape = (chain_n, draw_n)
            dims = ["chain", "draw"]

            # Design matrix without intercept
            X = self._design_matrix_without_intercept

            # Re-scale intercept for centered predictors
            common_terms = []
            for term in self.spec.common_terms.values():
                if term.alias:
                    common_terms += [term.alias]
                else:
                    common_terms += [term.name]

            if self.spec.response.coords:
                # Grab the first object in a dictionary
                levels = list(self.spec.response.coords.values())[0]
                shape += (len(levels),)
                dims += list(self.spec.response.coords)

            posterior = idata.posterior.stack(samples=dims)
            coefs = np.vstack([np.atleast_2d(posterior[name].values) for name in common_terms])

            if self.spec.intercept_term.alias:
                intercept_name = self.spec.intercept_term.alias
            else:
                intercept_name = self.spec.intercept_term.name

            idata.posterior[intercept_name] = idata.posterior[intercept_name] - np.dot(
                X.mean(0), coefs
            ).reshape(shape)

        if include_mean:
            self.spec.predict(idata)

        return idata

    def _run_vi(self, **kwargs):
        with self.model:
            self.vi_approx = pm.fit(**kwargs)
        return self.vi_approx

    def _run_laplace(self, draws, omit_offsets, include_mean):
        """Fit a model using a Laplace approximation.

        Mainly for pedagogical use, provides reasonable results for approximately
        Gaussian posteriors. The approximation can be very poor for some models
        like hierarchical ones. Use ``mcmc``, ``nuts_numpyro``, ``nuts_blackjax``
        or ``vi`` for better approximations.

        Parameters
        ----------
        draws: int
            The number of samples to draw from the posterior distribution.
        omit_offsets: bool
            Omits offset terms in the ``InferenceData`` object returned when the model includes
            group specific effects.
        include_mean: bool
            Compute the posterior of the mean response.

        Returns
        -------
        An ArviZ's InferenceData object.
        """
        with self.model:
            maps = pm.find_MAP()
            n_maps = deepcopy(maps)
            for m in maps:
                if pm.util.is_transformed_name(m):
                    n_maps.pop(pm.util.get_untransformed_name(m))

            hessian = pm.find_hessian(n_maps)

        if np.linalg.det(hessian) == 0:
            raise np.linalg.LinAlgError("Singular matrix. Use mcmc or vi method")

        cov = np.linalg.inv(hessian)
        modes = np.concatenate([np.atleast_1d(v) for v in n_maps.values()])

        samples = np.random.multivariate_normal(modes, cov, size=draws)

        idata = _posterior_samples_to_idata(samples, self.model)
        idata = self._clean_results(idata, omit_offsets, include_mean)
        return idata


def _posterior_samples_to_idata(samples, model):
    """Create InferenceData from samples.

    Parameters
    ----------
    samples: array
        Posterior samples
    model: PyMC model

    Returns
    -------
    An ArviZ's InferenceData object.
    """
    initial_point = model.initial_point(seed=None)
    variables = model.value_vars

    var_info = {}
    for name, value in initial_point.items():
        var_info[name] = (value.shape, value.size)

    length_pos = len(samples)
    varnames = [v.name for v in variables]

    with model:
        strace = pm.backends.ndarray.NDArray(name=model.name)  # pylint:disable=no-member
        strace.setup(length_pos, 0)
    for i in range(length_pos):
        value = []
        size = 0
        for varname in varnames:
            shape, new_size = var_info[varname]
            var_samples = samples[i][size : size + new_size]
            value.append(var_samples.reshape(shape))
            size += new_size
        strace.record(point=dict(zip(varnames, value)))

    idata = pm.to_inference_data(pm.backends.base.MultiTrace([strace]), model=model)
    return idata


def add_lkj(backend, terms, eta=1):
    """Add correlated prior for group-specific effects.

    This function receives a list of group-specific terms that share their `grouper`, constructs
    a multivariate Normal prior with LKJ prior on the correlation matrix, and adds the necessary
    variables to the model. It uses a non-centered parametrization.

    Parameters
    ----------
    terms: list
        A list of terms that share a common grouper (i.e. ``1|Group`` and ``Variable|Group`` in
        formula notation).
    eta: num
        The value for the eta parameter in the LKJ distribution.

    Parameters
    ----------
    mu
        The contribution to the linear predictor of the roup-specific terms in ``terms``.
    """

    # Parameters
    # grouper: The name of the grouper.build_group_specific_distribution
    # rows: Sum of the number of columns in all the "Xi" matrices for a given grouper.
    #       Same than the order of L
    # cols: Number of groups in the grouper variable
    mu = 0
    grouper = terms[0].name.split("|")[1]
    rows = int(np.sum([term.predictor.shape[1] for term in terms]))
    cols = int(terms[0].grouper.shape[1])  # not the most beautiful, but works

    # Construct sigma
    # Horizontally stack the sigma values for all the hyperpriors
    sigma = np.hstack([term.prior.args["sigma"].args["sigma"] for term in terms])

    # Reconstruct the hyperprior for the standard deviations, using one variable
    sigma = pm.HalfNormal.dist(sigma=sigma, shape=rows)

    # Obtain Cholesky factor for the covariance
    # pylint: disable=unused-variable, disable=unpacking-non-sequence
    (lkj_decomp, corr, sigma,) = pm.LKJCholeskyCov(
        "_LKJCholeskyCov_" + grouper,
        n=rows,
        eta=eta,
        sd_dist=sigma,
        compute_corr=True,
        store_in_trace=False,
    )

    coefs_offset = pm.Normal("_LKJ_" + grouper + "_offset", mu=0, sigma=1, shape=(rows, cols))
    coefs = at.dot(lkj_decomp, coefs_offset).T

    ## Separate group-specific terms
    start = 0
    for term in terms:
        label = term.name
        dims = list(term.coords)

        # Add coordinates to the model, only if they are not added yet.
        for name, values in term.coords.items():
            if name not in backend.model.coords:
                backend.model.add_coords({name: values})
        backend.coords.update(**term.coords)

        predictor = term.predictor.squeeze()
        delta = term.predictor.shape[1]

        if delta == 1:
            idx = start
        else:
            idx = slice(start, start + delta)

        # Add prior for the parameter
        coef = pm.Deterministic(label, coefs[:, idx], dims=dims)
        coef = coef[term.group_index]

        # Add standard deviation of the hyperprior distribution
        group_dim = [dim for dim in dims if dim.endswith("_group_expr")]
        pm.Deterministic(label + "_sigma", sigma[idx], dims=group_dim)

        # Account for the contribution of the term to the linear predictor
        if predictor.ndim > 1:
            for col in range(predictor.shape[1]):
                mu += coef[:, col] * predictor[:, col]
        else:
            mu += coef * predictor
        start += delta

    # TO DO: Add correlations
    return mu



import numpy as np
import pymc as pm
import aesara.tensor as at

from bambi.backend.utils import has_hyperprior, get_distribution
from bambi.families.multivariate import Categorical, Multinomial
from bambi.families.univariate import Beta, Binomial, Gamma
from bambi.priors import Prior


class CommonTerm:
    """Representation of a common effects term in PyMC

    An object that builds the PyMC distribution for a common effects term. It also contains the
    coordinates that we then add to the model.

    Parameters
    ----------
    term: bambi.terms.Term
        An object representing a common effects term.
    """

    def __init__(self, term):
        self.term = term
        self.coords = self.get_coords()

    def build(self, spec):
        data = self.term.data
        label = self.name
        dist = self.term.prior.name
        args = self.term.prior.args
        distribution = get_distribution(dist)

        # Dims of the response variable
        response_dims = []
        if isinstance(spec.family, (Categorical, Multinomial)):
            response_dims = list(spec.response.coords)
            response_dims_n = len(spec.response.coords[response_dims[0]])

            # Arguments may be of shape (a,) but we need them to be of shape (a, b)
            # a: length of predictor coordinates
            # b: length of response coordinates
            for key, value in args.items():
                if value.ndim == 1:
                    args[key] = np.hstack([value[:, np.newaxis]] * response_dims_n)

        dims = list(self.coords) + response_dims
        if dims:
            coef = distribution(label, dims=dims, **args)
        else:
            coef = distribution(label, shape=data.shape[1], **args)

        # Prepends one dimension if response is multivariate categorical and predictor is 1d
        if response_dims and len(dims) == 1:
            coef = coef[np.newaxis, :]

        return coef, data

    def get_coords(self):
        coords = {}
        if self.term.categorical:
            name = self.name + "_dim"
            coords[name] = self.term.term.levels
        # Not categorical but multi-column, like when we use splines
        elif self.term.data.shape[1] > 1:
            name = self.name + "_dim"
            coords[name] = list(range(self.term.data.shape[1]))
        return coords

    @property
    def name(self):
        if self.term.alias:
            return self.term.alias
        return self.term.name


class GroupSpecificTerm:
    """Representation of a group specific effects term in PyMC

    Creates an object that builds the PyMC distribution for a group specific effect. It also
    contains the coordinates that we then add to the model.

    Parameters
    ----------
    term: bambi.terms.GroupSpecificTerm
        An object representing a group specific effects term.
    noncentered: bool
        Specifies if we use non-centered parametrization of group-specific effects.
    """

    def __init__(self, term, noncentered):
        self.term = term
        self.noncentered = noncentered
        self.coords = self.get_coords()

    def build(self, spec):
        label = self.name
        dist = self.term.prior.name
        kwargs = self.term.prior.args
        predictor = self.term.predictor.squeeze()

        # Dims of the response variable (e.g. categorical)
        response_dims = []
        if isinstance(spec.family, (Categorical, Multinomial)):
            response_dims = list(spec.response.coords)

        dims = list(self.coords) + response_dims
        # Squeeze ensures we don't have a shape of (n, 1) when we mean (n, )
        # This happens with categorical predictors with two levels and intercept.
        coef = self.build_distribution(dist, label, dims=dims, **kwargs).squeeze()
        coef = coef[self.term.group_index]

        return coef, predictor

    def get_coords(self):
        coords = {}

        # Use the name of the alias if there's an alias
        if self.term.alias:
            expr, factor = self.term.alias, self.term.alias
        else:
            expr, factor = self.term.name.split("|")

        # The group is always a coordinate we add to the model.
        coords[factor + "__factor_dim"] = self.term.groups

        if self.term.categorical:
            levels = self.term.term.expr.levels
            coords[expr + "__expr_dim"] = levels
        elif self.term.predictor.ndim == 2 and self.term.predictor.shape[1] > 1:
            coords[expr + "__expr_dim"] = [str(i) for i in range(self.term.predictor.shape[1])]
        return coords

    def build_distribution(self, dist, label, **kwargs):
        """Build and return a PyMC Distribution."""
        dist = get_distribution(dist)

        if "dims" in kwargs:
            group_dim = [dim for dim in kwargs["dims"] if dim.endswith("__expr_dim")]
            kwargs = {
                k: self.expand_prior_args(k, v, label, dims=group_dim) for (k, v) in kwargs.items()
            }
        else:
            kwargs = {k: self.expand_prior_args(k, v, label) for (k, v) in kwargs.items()}

        if self.noncentered and has_hyperprior(kwargs):
            sigma = kwargs["sigma"]
            offset = pm.Normal(label + "_offset", mu=0, sigma=1, dims=kwargs["dims"])
            return pm.Deterministic(label, offset * sigma, dims=kwargs["dims"])
        return dist(label, **kwargs)

    def expand_prior_args(self, key, value, label, **kwargs):
        # kwargs are used to pass 'dims' for group specific terms.
        if isinstance(value, Prior):
            # If there's an alias for the hyperprior, use it.
            key = self.term.hyperprior_alias.get(key, key)
            return self.build_distribution(value.name, f"{label}_{key}", **value.args, **kwargs)
        return value

    @property
    def name(self):
        if self.term.alias:
            return self.term.alias
        return self.term.name


class InterceptTerm:
    """Representation of an intercept term in a PyMC model.

    Parameters
    ----------
    term: bambi.terms.Term
        An object representing the intercept. This has ``.kind == "intercept"``
    """

    def __init__(self, term):
        self.term = term

    def build(self, spec):
        dist = get_distribution(self.term.prior.name)
        label = self.name
        # Pre-pends one dimension if response is multi-categorical
        if isinstance(spec.family, (Categorical, Multinomial)):
            dims = list(spec.response.coords)
            dist = dist(label, dims=dims, **self.term.prior.args)[np.newaxis, :]
        else:
            # NOTE: Intercept only models with shape=1 don't work anymore
            #       It seems that 'shape=1' is not needed anymore?
            # dist = dist(label, shape=1, **self.term.prior.args)
            dist = dist(label, **self.term.prior.args)
        return dist

    @property
    def name(self):
        if self.term.alias:
            return self.term.alias
        return self.term.name


class ResponseTerm:
    """Representation of a response term in a PyMC model.

    Parameters
    ----------
    term : bambi.terms.ResponseTerm
        The response term as represented in Bambi.
    family : bambi.famlies.Family
        The model family.
    """

    def __init__(self, term, family):
        self.term = term
        self.family = family

    def build(self, nu, invlinks):
        """Create and return the response distribution for the PyMC model.

        nu : aesara.tensor.var.TensorVariable
            The linear predictor in the PyMC model.
        invlinks : dict
            A dictionary where names are names of inverse link functions and values are functions
            that can operate with Aesara tensors.
        """
        data = self.term.data.squeeze()

        # Take the inverse link function that maps from linear predictor to the mean of likelihood
        if self.family.link.name in invlinks:
            linkinv = invlinks[self.family.link.name]
        else:
            linkinv = self.family.link.linkinv_backend

        # Add column of zeros to the linear predictor for the reference level (the first one)
        if isinstance(self.family, (Categorical, Multinomial)):
            # Make sure intercept-only models work
            nu = np.ones((data.shape[0], 1)) * nu
            nu = at.concatenate([np.zeros((data.shape[0], 1)), nu], axis=1)

        # Add mean parameter and observed data
        kwargs = {self.family.likelihood.parent: linkinv(nu), "observed": data}

        # Add auxiliary parameters
        kwargs = self.build_auxiliary_parameters(kwargs)

        # Build the response distribution
        dist = self.build_response_distribution(kwargs)

        return dist

    def build_auxiliary_parameters(self, kwargs):
        # Build priors for the auxiliary parameters in the likelihood (e.g. sigma in Gaussian)
        if self.family.likelihood.priors:
            for key, value in self.family.likelihood.priors.items():

                # Use the alias if there's one
                if key in self.family.aliases:
                    label = self.family.aliases[key]
                else:
                    label = f"{self.name}_{key}"

                if isinstance(value, Prior):
                    dist = get_distribution(value.name)
                    kwargs[key] = dist(label, **value.args)
                else:
                    kwargs[key] = value
        return kwargs

    def build_response_distribution(self, kwargs):
        # Get likelihood distribution
        dist = get_distribution(self.family.likelihood.name)

        # Handle some special cases
        if isinstance(self.family, Beta):
            # Beta distribution in PyMC uses alpha and beta, but we have mu and kappa.
            # alpha = mu * kappa
            # beta = (1 - mu) * kappa
            alpha = kwargs["mu"] * kwargs["kappa"]
            beta = (1 - kwargs["mu"]) * kwargs["kappa"]
            return dist(self.name, alpha=alpha, beta=beta, observed=kwargs["observed"])

        if isinstance(self.family, Binomial):
            successes = kwargs["observed"][:, 0].squeeze()
            trials = kwargs["observed"][:, 1].squeeze()
            return dist(self.name, p=kwargs["p"], observed=successes, n=trials)

        if isinstance(self.family, Gamma):
            # Gamma distribution is specified using mu and sigma, but we request prior for alpha.
            # We build sigma from mu and alpha.
            sigma = kwargs["mu"] / (kwargs["alpha"] ** 0.5)
            return dist(self.name, mu=kwargs["mu"], sigma=sigma, observed=kwargs["observed"])

        if isinstance(self.family, Multinomial):
            n = kwargs["observed"].sum(axis=1)
            return dist(self.name, p=kwargs["p"], observed=kwargs["observed"], n=n)

        return dist(self.name, **kwargs)

    @property
    def name(self):
        if self.term.alias:
            return self.term.alias
        return self.term.name



"""Base IO code for datasets. Heavily influenced by Arviz's (and scikit-learn's) implementation."""
import hashlib
import itertools
import os
import shutil

from collections import namedtuple
from urllib.request import urlretrieve

import pandas as pd


FileMetadata = namedtuple("FileMetadata", ["filename", "url", "checksum", "description"])
DATASETS = {
    "my_data": FileMetadata(
        filename="my_data.csv",
        url="https://ndownloader.figshare.com/files/28850355",
        checksum="1bfcdd10d0848c1811e33e467c92734fb488406ef3f9b9aae16a57b258a30fac",
        description="""
Toy dataset with one response variable "y" and two covariates "x" and "z".
""",
    ),
    "adults": FileMetadata(
        filename="adults.csv",
        url="https://ndownloader.figshare.com/files/28870743",
        checksum="27a5270ba720125dfb24a7708cbee0218b2ead36248ae244813655d03320e43e",
        description="""
A sample with census data from 1994 in United States.
""",
    ),
    "ANES": FileMetadata(
        filename="ANES_2016_pilot.csv",
        url="https://ndownloader.figshare.com/files/28870740",
        checksum="3106beb6ded5a592ea0405d23b868bd8e74c259d7a7f5242c907555692905772",
        description="""
The ANES is a nationally representative, cross-sectional survey used extensively in political
science. This is a dataset from the 2016 pilot study, consisting of responses from 1200 voting-age
 U.S. citizens.
""",
    ),
    "ESCS": FileMetadata(
        filename="ESCS.csv",
        url="https://ndownloader.figshare.com/files/28870722",
        checksum="0195545797a4258de138a205a013a84022bbe23e7ff47782b179055c706300b6",
        description="""
A longitudinal study of hundreds of adults who completed dozens of different self-report and
behavioral measures over the course of 15 years. Among the behavioral measures is an index of
illegal drug use.
""",
    ),
    "carclaims": FileMetadata(
        filename="carclaims.csv",
        url="https://ndownloader.figshare.com/files/28870713",
        checksum="74924bf5f0a6e5aa5453d87845cea05e6b41bb2052cf6f096d7f102235ae5cdf",
        description="""
67856 insurance policies and 4624 (6.8%) claims in Australia between 2004 and 2005
""",
    ),
    "batting": FileMetadata(
        filename="Batting.csv",
        url="https://ndownloader.figshare.com/files/29749140",
        checksum="bbbc9459632c738a07bbe0877970a7bbd1f4c2448193979337fe5bc3a4ab0228",
        description="""
Baseball Databank is a compilation of historical baseball data in a convenient, tidy format,
distributed under Open Data terms by the Baseball Data Bank.
""",
    ),
    "cherry_blossoms": FileMetadata(
        filename="cherry_blossoms.csv",
        url="https://figshare.com/ndownloader/files/31072807",
        checksum="b859dd4f64c231c76ecb80b78f26da71e2f92698c50e0ceb93be0399dee24f51",
        description="""
Historical Series of Phenological data for Cherry Tree Flowering at Kyoto City. Extracted from
the `rethinking` library in R.
""",
    ),
    "sleepstudy": FileMetadata(
        filename="sleepstudy.csv",
        url="https://figshare.com/ndownloader/files/31181002",
        checksum="0a002bec8be2fa9d40dbbf3d5038e614d113a4fd5bf8813f6f4271c3d6294675",
        description="""
The average reaction time per day (in milliseconds) for subjects in a sleep deprivation study.
Days 0-1 were adaptation and training (T1/T2), day 2 was baseline (B); sleep deprivation started
after day 2.

Reaction
    Average reaction time (ms)

Days
    Number of days of sleep deprivation

Subject
    Subject number on which the observation was made
""",
    ),
    "periwinkles": FileMetadata(
        filename="periwinkles.csv",
        url="https://ndownloader.figshare.com/files/34446077",
        checksum="50da9791b7a66fbcc9ea4dd828dc7a3a66d5e067faf10f3bfd143af6c590923a",
        description="""Data for 31 periwinkles transplanted downshore as a function of the distance
        travelled by them after release.""",
    ),
    "admissions": FileMetadata(
        filename="admissions.csv",
        url="https://figshare.com/ndownloader/files/34757857",
        checksum="41e2312ca09d50e99c2db67fbabc78d215df6ce71eefe880df5e9310a9fa8397",
        description="""Admission into graduate school data. This dataset has a binary response
        variable called 'admit'. There are three predictor variables: 'gre', 'gpa' and 'rank'.""",
    ),
}


def get_data_home(data_home=None):
    """Return the path of the Bambi data dir.

    This folder is used to avoid downloading the data several times.

    By default the data dir is set to a folder named 'bambi_data' in the user home folder.
    Alternatively, it can be set by the ``"BAMBI_DATA"`` environment variable or programmatically by
    giving an explicit folder path. The ``"~"`` symbol is expanded to the user home folder. If the
    folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home: str
        The path to Bambi data dir.
    """
    if data_home is None:
        data_home = os.environ.get("BAMBI_DATA", os.path.join("~", "bambi_data"))
    data_home = os.path.expanduser(data_home)
    if not os.path.exists(data_home):
        os.makedirs(data_home)
    return data_home


def clear_data_home(data_home=None):
    """Delete all the content of the data home cache.

    Parameters
    ----------
    data_home: str
        The path to Bambi data dir. By default a folder named ``"bambi_data"`` in the user home
        folder.
    """
    data_home = get_data_home(data_home)
    shutil.rmtree(data_home)


def _sha256(path):
    """Calculate the sha256 hash of the file at path."""
    sha256hash = hashlib.sha256()
    chunk_size = 8192
    with open(path, "rb") as buff:
        while True:
            buffer = buff.read(chunk_size)
            if not buffer:
                break
            sha256hash.update(buffer)

    return sha256hash.hexdigest()


def load_data(dataset=None, data_home=None):
    """Load a dataset.

    Run with no parameters to get a list of all available data sets.

    The directory to save can also be set with the environment variable ``BAMBI_HOME``.
    The checksum of the dataset is checked against a hardcoded value to watch for data corruption.
    Run ``bmb.clear_data_home()`` to clear the data directory.

    Parameters
    ----------
    dataset: str
        Name of dataset to load.
    data_home: str, optional
        Where to save remote datasets

    Returns
    -------
    pandas.DataFrame
    """
    home_dir = get_data_home(data_home=data_home)

    if dataset in DATASETS:
        datafile = DATASETS[dataset]
        file_path = os.path.join(home_dir, datafile.filename)

        if not os.path.exists(file_path):
            urlretrieve(datafile.url, file_path)
            checksum = _sha256(file_path)
            if datafile.checksum != checksum:
                raise IOError(
                    f"{file_path} has an SHA256 checksum ({checksum}) differing from expected "
                    f"({datafile.checksum}), file may be corrupted. Run `bambi.clear_data_home()` "
                    "and try again, or please open an issue."
                )
        return pd.read_csv(file_path)
    else:
        if dataset is None:
            return _list_datasets(home_dir)
        else:
            raise ValueError(
                f"Dataset {dataset} not found! "
                f"The following are available:\n{_list_datasets(home_dir)}"
            )


def _list_datasets(home_dir):
    """Get a string representation of all available datasets with descriptions."""
    lines = []
    for filename, resource in itertools.chain(DATASETS.items()):
        file_path = os.path.join(home_dir, filename)
        if not os.path.exists(file_path):
            location = f"location: {resource.url}"
        else:
            location = f"location: {file_path}"
        lines.append(f"{filename}\n{'=' * len(filename)}\n{resource.description}\n{location}")

    return f"\n\n{10 * '-'}\n\n".join(lines)



# pylint: disable = protected-access
# pylint: disable = too-many-function-args
from statistics import mode

import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from matplotlib.lines import Line2D
from matplotlib.patches import Patch
from pandas.api.types import is_categorical_dtype, is_numeric_dtype, is_string_dtype

from bambi.utils import listify
from bambi.plots.utils import get_group_offset, get_unique_levels


def create_cap_data(model, covariates, grid_n=200, groups_n=5):
    """Create data for a Conditional Adjusted Predictions plot

    Parameters
    ----------
    model : bambi.Model
        An instance of a Bambi model
    covariates : list
        A sequence of one or two names of variables. The first variable is taken as the main
        variable. If present, the second variable is a grouping variable.
    grid_n : int, optional
        The number of points used to evaluate the main covariate. Defaults to 200.
    groups_n : int, optional
        The number of groups to create when the grouping variable is numeric. Groups are based on
        equally spaced points. Defaults to 5.

    Returns
    -------
    pandas.DataFrame
        The data for the Conditional Adjusted Predictions plot.

    Raises
    ------
    ValueError
        When the number of covariates is larger than 2.
        When either the main or the group covariates are not numeric or categoric.
    """
    data = model.data
    covariates = listify(covariates)

    if len(covariates) not in [1, 2]:
        raise ValueError(f"The number of covariates must be 1 or 2. It's {len(covariates)}.")

    main = covariates[0]

    # If available, take the name of the grouping variable
    if len(covariates) == 1:
        group = None
    else:
        group = covariates[1]

    # Obtain data for main variable
    data_main = data[main]
    if is_numeric_dtype(data_main):
        main_values = np.linspace(np.min(data_main), np.max(data_main), grid_n)
    elif is_string_dtype(data_main) or is_categorical_dtype(data_main):
        main_values = np.unique(data_main)
    else:
        raise ValueError("Main covariate must be numeric or categoric.")

    # If available, obtain groups for grouping variable
    if group:
        group_data = data[group]
        if is_string_dtype(group_data) or is_categorical_dtype(group_data):
            group_values = np.unique(group_data)
        elif is_numeric_dtype(group_data):
            group_values = np.quantile(group_data, np.linspace(0, 1, groups_n))
        else:
            raise ValueError("Group covariate must be numeric or categoric.")

        # Reshape accordingly
        group_n = len(group_values)
        main_n = len(main_values)
        main_values = np.tile(main_values, group_n)
        group_values = np.repeat(group_values, main_n)
        data_dict = {main: main_values, group: group_values}
    else:
        data_dict = {main: main_values}

    # Construct dictionary of terms that are in the model
    terms = {}
    if model._design.common:
        terms.update(model._design.common.terms)

    if model._design.group:
        terms.update(model._design.group.terms)

    # Get default values for each variable in the model
    for term in terms.values():
        if hasattr(term, "components"):
            for component in term.components:
                name = component.name
                if name not in data_dict:
                    # For numeric predictors, select the mean.
                    if component.kind == "numeric":
                        data_dict[name] = np.mean(data[name])
                    # For categoric predictors, select the most frequent level.
                    elif component.kind == "categoric":
                        data_dict[name] = mode(data[name])

    cap_data = pd.DataFrame(data_dict)

    # Make sure new types are same types than the original columns
    for column in cap_data:
        cap_data[column] = cap_data[column].astype(data[column].dtype)
    return cap_data


def plot_cap(model, idata, covariates, use_hdi=True, hdi_prob=None, legend=True, ax=None):
    """Plot Conditional Adjusted Predictions

    Parameters
    ----------
    model : bambi.Model
        The model for which we want to plot the predictions.
    idata : arviz.InferenceData
        The InferenceData object that contains the samples from the posterior distribution of
        the model.
    covariates : list
        A sequence of one or two names of variables. The first variable is taken as the main
        variable. If present, the second variable is a grouping variable.
    use_hdi : bool, optional
        Whether to compute the highest density interval (defaults to True) or the quantiles.
    hdi_prob : float, optional
        The probability for the credibility intervals. Must be between 0 and 1. Defaults to 0.94.
        Changing the global variable ``az.rcParam["stats.hdi_prob"]`` affects this default.
    legend : bool, optional
        Whether to automatically include a legend in the plot. Defaults to ``True``.
    ax : matplotlib.axes._subplots.AxesSubplot, optional
        A matplotlib axes object. If None, this function instantiates a new axes object.
        Defaults to ``None``.

    Returns
    -------
    matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        A tuple with the figure and the axes.

    Raises
    ------
    ValueError
        When ``level`` is not within 0 and 1.
        When the main covariate is not numeric or categoric.
    """

    covariates = listify(covariates)
    assert len(covariates) in [1, 2]

    cap_data = create_cap_data(model, covariates)
    idata = model.predict(idata, data=cap_data, inplace=False)

    if hdi_prob is None:
        hdi_prob = az.rcParams["stats.hdi_prob"]

    if not 0 < hdi_prob < 1:
        raise ValueError(f"'hdi_prob' must be greater than 0 and smaller than 1. It is {hdi_prob}.")

    y_hat = idata.posterior[f"{model.response.name}_mean"]
    y_hat_mean = y_hat.mean(("chain", "draw"))

    if use_hdi:
        y_hat_bounds = az.hdi(y_hat, hdi_prob)[f"{model.response.name}_mean"].T
    else:
        lower_bound = round((1 - hdi_prob) / 2, 4)
        upper_bound = 1 - lower_bound
        y_hat_bounds = y_hat.quantile(q=(lower_bound, upper_bound), dim=("chain", "draw"))

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    main = covariates[0]
    if is_numeric_dtype(cap_data[main]):
        ax = _plot_cap_numeric(covariates, cap_data, y_hat_mean, y_hat_bounds, legend, ax)
    elif is_categorical_dtype(cap_data[main]) or is_string_dtype(cap_data[main]):
        ax = _plot_cap_categoric(covariates, cap_data, y_hat_mean, y_hat_bounds, legend, ax)
    else:
        raise ValueError("Main covariate must be numeric or categoric.")

    ax.set(xlabel=main, ylabel=model.response.name)
    return fig, ax


def _plot_cap_numeric(covariates, cap_data, y_hat_mean, y_hat_bounds, legend, ax):
    if len(covariates) == 1:
        main = covariates[0]
        ax.plot(cap_data[main], y_hat_mean, solid_capstyle="butt")
        ax.fill_between(cap_data[main], y_hat_bounds[0], y_hat_bounds[1], alpha=0.5)
    else:
        main, group = covariates
        groups = get_unique_levels(cap_data[group])

        for i, grp in enumerate(groups):
            idx = (cap_data[group] == grp).values
            ax.plot(cap_data.loc[idx, main], y_hat_mean[idx], color=f"C{i}", solid_capstyle="butt")
            ax.fill_between(
                cap_data.loc[idx, main],
                y_hat_bounds[0][idx],
                y_hat_bounds[1][idx],
                alpha=0.3,
                color=f"C{i}",
            )

        if legend:
            handles = [
                (
                    Line2D([], [], color=f"C{i}", solid_capstyle="butt"),
                    Patch(color=f"C{i}", alpha=0.3, lw=1),
                )
                for i in range(len(groups))
            ]
            ax.legend(
                handles,
                tuple(groups),
                title=group,
                handlelength=1.3,
                handleheight=1,
                bbox_to_anchor=(1.03, 0.5),
                loc="center left",
            )

    return ax


def _plot_cap_categoric(covariates, cap_data, y_hat_mean, y_hat_bounds, legend, ax):
    main = covariates[0]
    main_levels = get_unique_levels(cap_data[main])
    main_levels_n = len(main_levels)
    idxs_main = np.arange(main_levels_n)

    if len(covariates) == 1:
        ax.scatter(idxs_main, y_hat_mean)
        ax.vlines(idxs_main, y_hat_bounds[0], y_hat_bounds[1])
    else:
        group = covariates[1]
        group_levels = get_unique_levels(cap_data[group])
        group_levels_n = len(group_levels)
        offset_bounds = get_group_offset(group_levels_n)
        offset_groups = np.linspace(-offset_bounds, offset_bounds, group_levels_n)

        for i, grp in enumerate(group_levels):
            idx = (cap_data[group] == grp).values
            idxs = idxs_main + offset_groups[i]
            ax.scatter(idxs, y_hat_mean[idx], color=f"C{i}")
            ax.vlines(idxs, y_hat_bounds[0][idx], y_hat_bounds[1][idx], color=f"C{i}")

        if legend:
            handles = [
                Line2D([], [], c=f"C{i}", marker="o", label=level)
                for i, level in enumerate(group_levels)
            ]
            ax.legend(handles=handles, title=group, bbox_to_anchor=(1.03, 0.5), loc="center left")

    ax.set_xticks(idxs_main)
    ax.set_xticklabels(main_levels)
    return ax



import numpy as np


def get_unique_levels(x):
    if hasattr(x, "dtype") and hasattr(x.dtype, "categories"):
        levels = list(x.dtype.categories)
    else:
        levels = np.unique(x)
    return levels


def get_group_offset(n, lower=0.05, upper=0.4):
    # Complementary log log function, scaled.
    # See following code to have an idea of how this function looks like
    # lower, upper = 0.05, 0.4
    # x = np.linspace(2, 9)
    # y = get_group_offset(x, lower, upper)
    # fig, ax = plt.subplots(figsize=(8, 5))
    # ax.plot(x, y)
    # ax.axvline(2, color="k", ls="--")
    # ax.axhline(lower, color="k", ls="--")
    # ax.axhline(upper, color="k", ls="--")
    intercept, slope = 3.25, 1
    return lower + np.exp(-np.exp(intercept - slope * n)) * (upper - lower)



from bambi.families import Likelihood
from bambi.priors import Prior

from bambi.families.univariate import (
    Bernoulli,
    Beta,
    Binomial,
    Gamma,
    Gaussian,
    NegativeBinomial,
    Laplace,
    Poisson,
    StudentT,
    VonMises,
    Wald,
)

from bambi.families.multivariate import Categorical, Multinomial

# Default parameters for PyMC distributions
SETTINGS_DISTRIBUTIONS = {
    "Bernoulli": {"p": 0.5},
    "Beta": {"alpha": 1, "beta": 1},
    "Binomial": {"n": 1, "p": 0.5},
    "Cauchy": {"alpha": 0, "beta": 1},
    "Flat": {},
    "Gamma": {"alpha": 2, "beta": 0.1},
    "HalfCauchy": {"beta": 1},
    "HalfFlat": {},
    "HalfNormal": {"sigma": 1},
    "Normal": {"mu": 0, "sigma": 1},
    "NegativeBinomial": {"alpha": 1, "mu": 1},
    "Laplace": {"mu": 0, "b": 1},
    "Poisson": {"mu": 1},
    "StudentT": {"lam": 1, "nu": 1},
    "VonMises": {"mu": 0, "kappa": 1},
    "Wald": {"mu": 1, "lam": 1},
}

# fmt: off
BUILTIN_FAMILIES = {
    "bernoulli": {
        "likelihood": {
            "name": "Bernoulli",
            "args": {},
            "parent": "p",
        },
        "link": "logit",
        "family": Bernoulli,
    },
    "beta": {
        "likelihood": {
            "name": "Beta",
            "args": {
                "kappa": "HalfCauchy"
            },
            "parent": "mu",
        },
        "link": "logit",
        "family": Beta,
    },
    "binomial": {
        "likelihood": {
            "name": "Binomial",
            "args": {},
            "parent": "p",
        },
        "link": "logit",
        "family": Binomial,
    },
    "categorical": {
        "likelihood": {
            "name": "Categorical",
            "args": {},
            "parent": "p",
        },
        "link": "softmax",
        "family": Categorical,
    },
    "gamma": {
        "likelihood": {
            "name": "Gamma",
            "args": {
                "alpha": "HalfCauchy"
            },
            "parent": "mu",
        },
        "link": "inverse",
        "family": Gamma,
    },
    "gaussian": {
        "likelihood": {
            "name": "Normal",
            "args": {
                "sigma": "HalfNormal"
            },
            "parent": "mu",
        },
        "link": "identity",
        "family": Gaussian,
    },
    "multinomial": {
        "likelihood": {
            "name": "Multinomial",
            "args": {},
            "parent": "p"
        },
        "link": "softmax",
        "family": Multinomial,
    },
    "negativebinomial": {
        "likelihood": {
            "name": "NegativeBinomial",
            "args": {
                "alpha": "HalfCauchy"
            },
            "parent": "mu",
        },
        "link": "log",
        "family": NegativeBinomial,
    },
    "laplace": {
        "likelihood": {
            "name": "Laplace",
            "args": {
                "b": "HalfNormal"
            },
            "parent": "mu",
        },
        "link": "identity",
        "family": Laplace,
    },
    "poisson": {
        "likelihood": {
            "name": "Poisson",
            "args": {},
            "parent": "mu",
        },
        "link": "log",
        "family": Poisson,
    },
    "t": {
        "likelihood": {
            "name": "StudentT",
            "args": {
                "sigma": "HalfNormal",
                "nu": "Gamma"
            },
            "parent": "mu",
        },
        "link": "identity",
        "family": StudentT,
    },
    "vonmises": {
        "likelihood": {
            "name": "VonMises",
            "args": {
                "kappa": "HalfNormal"
            },
            "parent": "mu",
        },
        "link": "tan_2",
        "family": VonMises,
    },
    "wald": {
        "likelihood": {
            "name": "Wald",
            "args": {
                "lam": "HalfCauchy"
            },
            "parent": "mu",
        },
        "link": "inverse_squared",
        "family": Wald,
    },
}
# fmt: on


def generate_prior(dist, **kwargs):
    """Generate a Prior distribution.

    The parameter ``kwargs`` is used to pass hyperpriors that are assigned to the parameters of
    the prior to be built.

    Parameters
    ----------
    dist: str, int, float
        If a string, it is the name of the prior distribution with default values taken from
        ``SETTINGS_DISTRIBUTIONS``. If a number, it is a factor used to scale the standard deviation
        of the priors generated automatically by Bambi.

    Raises
    ------
    ValueError
        If ``dist`` is not a string or a number.

    Returns
    -------
    Prior
        The Prior instance.
    """
    if isinstance(dist, str):
        prior = Prior(dist, **SETTINGS_DISTRIBUTIONS[dist])
        if kwargs:
            prior.update(**{k: generate_prior(v) for k, v in kwargs.items()})
    elif isinstance(dist, (int, float)):
        prior = dist
    else:
        raise ValueError("'dist' must be the name of a distribution or a numeric value.")
    return prior


def generate_likelihood(name, args, parent):
    """Generate a Likelihood instance.

    Parameters
    ----------
    name: str
        The name of the likelihood function.
    args: dict
        Indicates the auxiliary parameters and the values for their default priors. The keys are the
        names of the parameters and the values are passed to ``generate_prior()`` to obtain the
        actual instance of ``bambi.Prior``.
    parent: str
        The name of the parent parameter. In other words, the name of the mean parameter in the
        likelihood function.

    Returns
    -------
    bambi.Likelihood
        The likelihood instance.
    """
    priors = {k: generate_prior(v) for k, v in args.items()}
    return Likelihood(name, parent, **priors)


def generate_family(name, likelihood, link, family):
    """Generate a Bambi family.

    Parameters
    ----------
    name: str
        The name of the family.
    likelihood: bambi.Likelihood
        A representation of the likelihood function that corresponds to the family being created.
    link: bambi.Link
        A representation of the link function that corresponds to the family being created.
    family: subclass of bambi.Family
        A subclass of bambi.Family that generates the instance of the desired family.

    Returns
    -------
    bambi.Family
        The family instance.
    """
    likelihood = generate_likelihood(**likelihood)
    return family(name, likelihood, link)


def get_default_prior(term_type):
    """Generate a Prior based on the default settings

    The following summarises default priors for each type of term:

    * intercept: Normal prior.
    * common: Normal prior.
    * intercept_flat: Uniform prior.
    * common_flat: Uniform prior.
    * group_specific: Normal prior where its sigma has a HalfNormal hyperprior.
    * group_specific_flat: Normal prior where its sigma has a HalfFlat hyperprior.

    Parameters
    ----------
    term_type: str
        The type of the term for which the default prior is wanted.

    Raises
    ------
    ValueError
        If ``term_type`` is not within the values listed above.

    Returns
    -------
    prior: Prior
        The instance of Prior according to the ``term_type``.
    """
    if term_type in ["intercept", "common"]:
        prior = generate_prior("Normal")
    elif term_type in ["intercept_flat", "common_flat"]:
        prior = generate_prior("Flat")
    elif term_type == "group_specific":
        prior = generate_prior("Normal", sigma="HalfNormal")
    elif term_type == "group_specific_flat":
        prior = generate_prior("Normal", sigma="HalfFlat")
    else:
        raise ValueError("Unrecognized term type.")
    return prior


def get_builtin_family(name):
    """Generate a built-in ``bambi.families.Family`` instance.

    Given the name of a built-in family, this function returns a ``bambi.families.Family`` instance
    that is constructed by calling other utility functions that construct the
    ``bambi.families.Likelihood`` and the ``bambi.priors.Prior`` instances that are needed to build
    the family.

    The available built-in families are found in ``SETTINGS_FAMILIES``.

    Parameters
    ----------
    name: str
        The name of the built-in family.

    Raises
    ------
    ValueError
        If ``name`` is not the name of a built-in family.

    Returns
    -------
    bambi.families.Family
        The family instance.
    """
    if name in BUILTIN_FAMILIES:
        return generate_family(name, **BUILTIN_FAMILIES[name])
    raise ValueError(f"'{name}' is not a valid built-in family name.")



import numpy as np


class Prior:
    """Abstract specification of a term prior.

    Parameters
    ----------
    name : str
        Name of prior distribution. Must be the name of a PyMC distribution
        (e.g., ``"Normal"``, ``"Bernoulli"``, etc.)
    auto_scale: bool
        Whether to adjust the parameters of the prior or use them as passed. Default to ``True``.
    kwargs : dict
        Optional keywords specifying the parameters of the named distribution.
    """

    def __init__(self, name, auto_scale=True, **kwargs):
        self.name = name
        self.auto_scale = auto_scale
        self.args = {}
        self.update(**kwargs)

    def update(self, **kwargs):
        """Update the arguments of the prior with additional arguments.

        Parameters
        ----------
        kwargs : dict
            Optional keyword arguments to add to prior args.
        """
        # The backend expect numpy arrays, so make sure all numeric values are represented as such.
        kwargs_ = {}
        for key, val in kwargs.items():
            if isinstance(val, (int, float)):
                val = np.array(val)
            elif isinstance(val, np.ndarray):
                val = val.squeeze()
            kwargs_[key] = val
        self.args.update(kwargs_)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __str__(self):
        args = ", ".join(
            [
                f"{k}: {np.round_(v, 4)}" if not isinstance(v, type(self)) else f"{k}: {v}"
                for k, v in self.args.items()
            ]
        )
        return f"{self.name}({args})"

    def __repr__(self):
        return self.__str__()



import os
import sys

from setuptools import find_packages, setup

PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))
README_FILE = os.path.join(PROJECT_ROOT, "README.md")
VERSION_FILE = os.path.join(PROJECT_ROOT, "bambi", "version.py")
REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements.txt")
OPTIONAL_REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements-optional.txt")
MINIMUM_PYTHON_VERSION = (3, 7, 2)


def check_installation():
    if sys.version_info < MINIMUM_PYTHON_VERSION:
        version = ".".join(str(i) for i in MINIMUM_PYTHON_VERSION)
        sys.stderr.write(
            f"[{sys.argv[0]}] - Error: Your Python interpreter must be {version} or greater.\n"
        )
        sys.exit(-1)


def get_long_description():
    with open(README_FILE, encoding="utf-8") as buff:
        return buff.read()


def get_requirements():
    with open(REQUIREMENTS_FILE, encoding="utf-8") as buff:
        return buff.read().splitlines()


def get_optional_requirements():
    with open(OPTIONAL_REQUIREMENTS_FILE, encoding="utf-8") as buff:
        return buff.read().splitlines()


def get_version():
    with open(VERSION_FILE, encoding="utf-8") as buff:
        exec(buff.read())  # pylint: disable=exec-used
    return vars()["__version__"]


check_installation()

__version__ = get_version()


setup(
    name="bambi",
    version=__version__,
    description="BAyesian Model Building Interface in Python",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="http://github.com/bambinos/bambi",
    download_url="https://github.com/bambinos/bambi/archive/%s.tar.gz" % __version__,
    install_requires=get_requirements(),
    extras_require={
        "jax": [get_optional_requirements()],
    },
    maintainer="Tomas Capretto",
    maintainer_email="tomicapretto@gmail.com",
    packages=find_packages(exclude=["tests", "test_*"]),
    package_data={"bambi": ["priors/config/*"]},
    license="MIT",
)



# pylint: disable=unused-argument
import numpy as np
from scipy import stats

from .family import Family


class UnivariateFamily(Family):
    def predict(self, model, posterior, linear_predictor):
        """Predict mean response"""
        mean = self.link.linkinv(linear_predictor)
        obs_n = mean.shape[-1]
        response_var = model.response.name + "_mean"
        response_dim = model.response.name + "_obs"

        # Drop var/dim if already present
        if response_var in posterior.data_vars:
            posterior = posterior.drop_vars(response_var)

        if response_dim in posterior.dims:
            posterior = posterior.drop_dims(response_dim)

        dims = ("chain", "draw", response_dim)
        posterior[response_var] = (dims, mean)
        posterior = posterior.assign_coords({response_dim: list(range(obs_n))})
        return posterior


class Bernoulli(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "logit", "probit", "cloglog"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        "Sample from posterior predictive distribution"
        mean = self.link.linkinv(linear_predictor)
        return np.random.binomial(1, mean)


class Beta(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "logit", "probit", "cloglog"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        kappa = posterior[model.response.name + "_kappa"].values
        kappa = kappa[:, :, np.newaxis]
        alpha = mean * kappa
        beta = (1 - mean) * kappa
        return np.random.beta(alpha, beta)


class Binomial(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "logit", "probit", "cloglog"]

    def posterior_predictive(self, model, posterior, linear_predictor, trials=None):
        if trials is None:
            trials = model.response.data[:, 1]
        mean = self.link.linkinv(linear_predictor)
        return np.random.binomial(trials.squeeze(), mean)


class Gamma(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log", "inverse"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        alpha = posterior[model.response.name + "_alpha"].values
        alpha = alpha[:, :, np.newaxis]
        beta = alpha / mean
        return np.random.gamma(alpha, 1 / beta)


class Gaussian(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log", "inverse"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        "Sample from posterior predictive distribution"
        mean = self.link.linkinv(linear_predictor)
        sigma = posterior[model.response.name + "_sigma"].values
        sigma = sigma[:, :, np.newaxis]
        return np.random.normal(mean, sigma)


class NegativeBinomial(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log", "cloglog"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        n = posterior[model.response.name + "_alpha"].values
        n = n[:, :, np.newaxis]
        p = n / (mean + n)
        return np.random.negative_binomial(n, p)


class Laplace(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log", "inverse"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        "Sample from posterior predictive distribution"
        mean = self.link.linkinv(linear_predictor)
        b = posterior[model.response.name + "_b"].values
        b = b[:, :, np.newaxis]
        return np.random.laplace(mean, b)


class Poisson(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        return np.random.poisson(mean)


class StudentT(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "log", "inverse"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        sigma = posterior[model.response.name + "_sigma"].values
        sigma = sigma[:, :, np.newaxis]

        if isinstance(self.likelihood.priors["nu"], (int, float)):
            nu = self.likelihood.priors["nu"]
        else:
            nu = posterior[model.response.name + "_nu"].values[:, :, np.newaxis]

        return stats.t.rvs(nu, mean, sigma)


class VonMises(UnivariateFamily):
    SUPPORTED_LINKS = ["identity", "tan_2"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        kappa = posterior[model.response.name + "_kappa"].values
        kappa = kappa[:, :, np.newaxis]
        return np.random.vonmises(mean, kappa)


class Wald(UnivariateFamily):
    SUPPORTED_LINKS = ["inverse", "inverse_squared", "identity", "log"]

    def posterior_predictive(self, model, posterior, linear_predictor):
        mean = self.link.linkinv(linear_predictor)
        lam = posterior[model.response.name + "_lam"].values
        lam = lam[:, :, np.newaxis]
        return np.random.wald(mean, lam)



import ast

import numpy as np


def listify(obj):
    """Wrap all non-list or tuple objects in a list.

    Provides a simple way to accept flexible arguments.
    """
    if obj is None:
        return []
    else:
        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]


def spacify(string, n=2):
    """Add spaces to the beginning of each line in a multi-line string."""
    space = n * " "
    return space + space.join(string.splitlines(True))


def multilinify(sequence, sep=","):
    """Make a multi-line string out of a sequence of strings."""
    sep += "\n"
    return "\n" + sep.join(sequence)


def c(*args):  # pylint: disable=invalid-name
    """Concatenate columns into a 2D NumPy Array"""
    return np.column_stack(args)


def extract_argument_names(expr, accepted_funcs):
    """Extract the names of the arguments passed to a function.

    This is used to extract the labels from function calls such as `c(y1, y2, y3, y3)`.

    Parameters
    ----------
    expr : str
        An expression that is parsed to extract the components of the call.
    accepted_funcs : list
        A list with the names of the functions that we accept to parse.

    Returns
    -------
    list
        If all criteria are met, the names of the arguments. Otherwise it returns None.
    """
    # Extract the first thing in the body
    parsed_expr = ast.parse(expr).body[0]
    if not isinstance(parsed_expr, ast.Expr):
        return None

    # Check the value is a call
    value = parsed_expr.value
    if not isinstance(value, ast.Call):
        return None

    # Check call name is the name of an exepcted function
    if value.func.id not in accepted_funcs:
        return None

    # Check all arguments are either names or constants
    args = value.args
    if not all(isinstance(arg, ast.Name) for arg in args):
        return None

    # We can safely build labels now
    labels = [arg.id for arg in args]

    if labels:
        return labels
    return None


extra_namespace = {"c": c}



from bambi.families.link import Link


class Family:
    """A specification of model family.

    Parameters
    ----------
    name: str
        The name of the family. It can be any string.
    likelihood: Likelihood
        A ``bambi.families.Likelihood`` instance specifying the model likelihood function.
    link: str or Link
        The name of the link function or a ``bambi.families.Link`` instance. The link function
        transforms the linear model prediction to the mean parameter of the likelihood function.

    Examples
    --------

    >>> import bambi as bmb

    Replicate the Gaussian built-in family.

    >>> sigma_prior = bmb.Prior("HalfNormal", sigma=1)
    >>> likelihood = bmb.Likelihood("Gaussian", parent="mu", sigma=sigma_prior)
    >>> family = bmb.Family("gaussian", likelihood, "identity")
    >>> # Then you can do
    >>> # bmb.Model("y ~ x", data, family=family)

    Replicate the Bernoulli built-in family.

    >>> likelihood = bmb.Likelihood("Bernoulli", parent="p")
    >>> family = bmb.Family("bernoulli", likelihood, "logit")
    """

    SUPPORTED_LINKS = [
        "cloglog",
        "identity",
        "inverse_squared",
        "inverse",
        "log",
        "logit",
        "probit",
        "softmax",
        "tan_2",
    ]

    def __init__(self, name, likelihood, link):
        self.name = name
        self.likelihood = likelihood
        self.link = link
        self.aliases = {}

    @property
    def link(self):
        return self._link

    @link.setter
    def link(self, x):
        if isinstance(x, str):
            self.check_string_link(x)
            self._link = Link(x)
        elif isinstance(x, Link):
            self._link = x
        else:
            raise ValueError(".link must be set to a string or a Link instance.")

    def check_string_link(self, link):
        if not link in self.SUPPORTED_LINKS:
            raise ValueError(f"Link '{link}' cannot be used with family '{self.name}'")

    def set_alias(self, name, alias):
        """Set alias for an auxiliary variable of the family

        Parameters
        ----------
        name: str
            The name of the variable
        alias: str
            The new name for the variable
        """
        self.aliases.update({name: alias})

    def __str__(self):
        msg_list = [f"Response distribution: {self.likelihood.name}", f"Link: {self.link.name}"]
        if self.likelihood.priors:
            priors_msg = "\n  ".join([f"{k} ~ {v}" for k, v in self.likelihood.priors.items()])
            msg_list += [f"Priors:\n  {priors_msg}"]
        msg = "\n".join(msg_list)
        return msg

    def __repr__(self):
        return self.__str__()



import numpy as np

from bambi.families.multivariate import Categorical, Multinomial
from bambi.families.univariate import Bernoulli
from bambi.utils import extract_argument_names, extra_namespace


class ResponseTerm:
    """Representation of a single response model term.

    Parameters
    ----------
    term : formulae.ResponseMatrix
        An object describing the response of the model,
        as returned by ``formulae.design_matrices().response``
    spec : bambi.Model
        The model where this response term is used.
    """

    def __init__(self, term, spec):
        self.name = term.name
        self.categorical = term.kind == "categoric"
        self.reference = None
        self.levels = None  # Not None for categorical variables
        self.binary = None  # Not None for categorical variables (either True or False)
        self.success = None  # Not None for binary variables (either True or False)
        self.alias = None
        self.data = None

        if self.categorical:
            if term.levels is None:
                self.binary = True
            else:
                self.levels = term.levels
                self.binary = len(term.levels) == 2

            if self.binary:
                self.success = get_success_level(term.term.term)
                if term.design_matrix.ndim == 1:
                    self.data = term.design_matrix
                else:
                    idx = self.levels.index(self.success)
                    self.data = term.design_matrix[:, idx]
            # Applies to the categorical family
            else:
                self.reference = get_reference_level(term.term.term)
                self.data = np.nonzero(term.design_matrix)[1]
        elif isinstance(spec.family, Bernoulli):
            # We've already checked the values are all 0 and 1
            self.success = 1
            self.data = term.design_matrix
        else:
            self.data = term.design_matrix

        # We use pymc coords when the response is multi-categorical.
        # These help to give the appropriate shape to coefficients and make the resulting
        # InferenceData object much cleaner
        self.coords = {}
        if isinstance(spec.family, Categorical):
            name = self.name + "_dim"
            self.coords[name] = [level for level in term.levels if level != self.reference]
        elif isinstance(spec.family, Multinomial):
            name = self.name + "_dim"
            labels = extract_argument_names(self.name, list(extra_namespace))
            if labels:
                self.levels = labels
            else:
                self.levels = [str(level) for level in range(self.data.shape[1])]
            labels = self.levels[1:]
            self.coords[name] = labels
        # TBD: Continue here when we add general multivariate responses.

    def set_alias(self, value):
        self.alias = value

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"shape: {self.data.squeeze().shape}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]
            if self.binary:
                args += [f"success: {self.success}"]
            else:
                args += [f"reference: {self.reference}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class Term:
    """Representation of a single (common) model term.

    Parameters
    ----------
    name: str
        Name of the term.
    term_dict: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().common.terms_info``
    data: ndarray
        The term values.
    prior: Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = False

    def __init__(self, name, term, data, prior=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.kind = term.kind
        self.levels = term.labels
        self.categorical = False
        self.term = term
        self.alias = None

        # If the term has one component, it's categorical if the component is categorical.
        # If the term has more than one component (i.e. it is an interaction), it's categorical if
        # at least one of the components is categorical.
        if self.kind == "interaction":
            if any(component.kind == "categoric" for component in term.components):
                self.categorical = True
        else:
            self.categorical = self.kind == "categoric"

        # Flag constant terms
        if self.categorical and len(term.levels) == 1 and (data == data[0]).all():
            raise ValueError(f"The term '{name}' has only 1 category!")

        if not self.categorical and self.kind != "intercept" and np.all(data == data[0]):
            raise ValueError(f"The term '{name}' is constant!")

        # Flag cell-means terms (i.e., full-rank coding), which receive special priors
        # To flag intercepts we use `self.kind`
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Obtain pymc coordinates, only for categorical components of a term.
        # A categorical component can have up to two coordinates if it is including with both
        # reduced and full rank encodings.
        self.coords = {}
        if self.categorical:
            name = self.name + "_dim"
            self.coords[name] = term.levels
        elif self.data.ndim > 1 and self.data.shape[1] > 1:
            name = self.name + "_dim"
            self.coords[name] = np.arange(self.data.shape[1])

    def set_alias(self, value):
        self.alias = value

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"kind: {self.kind}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class GroupSpecificTerm:
    # pylint: disable=too-many-instance-attributes
    """Representation of a single (group specific) model term.

    Parameters
    ----------
    name: str
        Name of the term.
    term: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().group.terms_info``
    data: (DataFrame, Series, ndarray)
        The term values.
    prior: Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = True

    def __init__(self, name, term, data, prior=None):
        self.categorical = False
        self.alias = None
        self.hyperprior_alias = {}

        self.name = name
        self.data = data
        self.prior = prior
        self.kind = term.kind
        self.groups = term.groups
        self.levels = term.labels
        self.grouper = term.factor.data
        self.predictor = term.expr.data
        self.group_index = self.invert_dummies(self.grouper)
        self.term = term

        # Determine if the expression is categorical
        if self.kind == "interaction":
            if any(component.kind == "categoric" for component in term.expr.components):
                self.categorical = True
        else:
            self.categorical = self.kind == "categoric"

        # Determine if the term represents cell-means encoding.
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Used in pymc model coords to label coordinates appropriately
        self.coords = {}

        # Group is always a coordinate added to the model.
        expr, factor = self.name.split("|")
        self.coords[factor + "__factor_dim"] = self.groups

        if self.categorical:
            self.coords[expr + "__expr_dim"] = term.expr.levels
        elif self.predictor.ndim == 2 and self.predictor.shape[1] > 1:
            self.coords[expr + "__expr_dim"] = [str(i) for i in range(self.predictor.shape[1])]

    def invert_dummies(self, dummies):
        """
        For the sake of computational efficiency (i.e., to avoid lots of large matrix
        multiplications in the backend), invert the dummy-coding process and represent full-rank
        dummies as a vector of indices into the coefficients.
        """
        vec = np.zeros(len(dummies), dtype=int)
        for i in range(1, dummies.shape[1]):
            vec[dummies[:, i] == 1] = i
        return vec

    def set_alias(self, value):
        self.alias = value

    def set_hyperprior_alias(self, name, value):
        self.hyperprior_alias.update({name: value})

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"groups: {self.groups}",
            f"type: {self.kind}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class OffsetTerm:
    """Representation of a single offset  term.

    Parameters
    ----------
    name: str
        Name of the term.
    term: formulae.terms.terms.Term
        A model term created in formulae.
    data: (DataFrame, Series, ndarray)
        The term values.
    """

    group_specific = False

    def __init__(self, name, term, data):
        self.name = name
        self.data = data.squeeze()
        self.kind = "offset"
        self.term = term
        self.alias = None
        self.coords = {}

    def set_alias(self, value):
        self.alias = value

    def __str__(self):
        args = [f"name: {self.name}", f"shape: {self.data.shape}"]
        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"
        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


# pylint: disable = protected-access
def get_reference_level(term):
    if term.kind != "categoric":
        return None

    if term.levels is None:
        return None

    levels = term.levels
    intermediate_data = term.components[0]._intermediate_data
    if hasattr(intermediate_data, "_contrast"):
        return intermediate_data._contrast.reference

    return levels[0]


# pylint: disable = protected-access
def get_success_level(term):
    if term.kind != "categoric":
        return None

    if term.levels is None:
        return term.components[0].reference

    levels = term.levels
    intermediate_data = term.components[0]._intermediate_data
    if hasattr(intermediate_data, "_contrast"):
        return intermediate_data._contrast.reference

    return levels[0]
